import json

from pathlib import Path
from subprocess import run


CRISPR_ARRAYS = [
    {
        'repeat_consensus': 'CGGTTTATCCCCGCTGGCGCGGGGAACACA',
        'spacers': [
            'AACCGAAACACACGATCAATCCGAATATGAG',
            'TTGGTGACAGTTTTTGTCACTGTTTTGGTGA',
            'CTAAGCATACATATCTGTTTTTAAACA'
        ],
        'repeats': 3
    }
]


def test_crispr_arrays(tmpdir):
    proc = run(
        [
            'bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--prefix', 'test',
            '--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-cds', '--skip-sorf', '--skip-ori', '--skip-gap', '--skip-plot',
            'test/data/GCF_000008865.2.fna.gz'
        ]
    )
    assert proc.returncode == 0

    results_path = Path(tmpdir).joinpath('test.json')
    assert Path.exists(results_path)
    
    results = None
    with results_path.open() as fh:
        results = json.load(fh)
    assert results is not None
    
    crispr_arrays = [feat for feat in results['features'] if feat['type'] == 'crispr']
    assert len(crispr_arrays) == 1

    for idx, crispr_array in enumerate(crispr_arrays):
        assert crispr_array['repeat_consensus'] == CRISPR_ARRAYS[idx]['repeat_consensus']
        assert len(crispr_array['repeats']) == CRISPR_ARRAYS[idx]['repeats']



import logging
import subprocess as sp

from collections import OrderedDict
from pathlib import Path
from typing import Sequence

import bakta.config as cfg
import bakta.constants as bc
import bakta.utils as bu


HIT_COVERAGE = 0.8
HIT_IDENTITY = 0.8
HIT_EVALUE = '1E-5'


log = logging.getLogger('ORI')


def predict_oris(data: dict, sequences_path: Path, ori_type: str) -> Sequence[dict]:
    """Search for oriT/C sequences."""

    database = 'oric.fna' if ori_type == bc.FEATURE_ORIC else 'orit.fna'
    output_path = cfg.tmp_path.joinpath('ori.blastn.tsv')
    cmd = [
        'blastn',
        '-query', str(cfg.db_path.joinpath(database)),
        '-subject', str(sequences_path),
        '-culling_limit', '1',
        '-evalue', HIT_EVALUE,
        '-num_threads', str(cfg.threads),
        '-outfmt', '6 qseqid qstart qend qlen sseqid sstart send length nident sstrand',
        '-out', str(output_path)
    ]
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('%s failed! blastn-error-code=%d', ori_type, proc.returncode)
        raise Exception(f'blastn error! error code: {proc.returncode}')

    # parse raw hits
    hits = {}
    with output_path.open() as fh:
        for line in fh:
            cols = line.strip().split('\t')
            hit = {
                'ori_id': cols[0],
                'ori_start': int(cols[1]),
                'ori_end': int(cols[2]),
                'ori_length': int(cols[3]),
                'sequence': cols[4],
                'sequence_start': int(cols[5]),
                'sequence_stop': int(cols[6]),
                'strand': bc.STRAND_FORWARD if cols[9] == 'plus' else bc.STRAND_REVERSE,
                'coverage': int(cols[7]) / int(cols[3]),
                'identity': int(cols[8]) / int(cols[7])
            }
            if(hit['strand'] == bc.STRAND_REVERSE):
                hit['sequence_start'], hit['sequence_stop'] = hit['sequence_stop'], hit['sequence_start']
            if(hit['coverage'] >= HIT_COVERAGE and hit['identity'] >= HIT_IDENTITY):
                sequence_hits = hits.get(hit['sequence'], [])
                sequence_hits.append(hit)
                if(len(sequence_hits) == 1):
                    hits[hit['sequence']] = sequence_hits
                log.debug(
                    'raw hit: type=%s, seq=%s, start=%i, stop=%i, strand=%s, coverage=%0.3f, identity=%0.3f',
                    ori_type, hit['sequence'], hit['sequence_start'], hit['sequence_stop'], hit['strand'], hit['coverage'], hit['identity']
                )

    # combine overlapping hits (simple 1D array peak detection)
    oris = []
    for seq in data['sequences']:
        sequence_hits = hits.get(seq['id'], None)
        if(sequence_hits):
            region_hits = [0] * (seq['length'] + 1)  # init with extra leading slot (start at 1)
            for hit in sequence_hits:
                for i in range(hit['sequence_start'], hit['sequence_stop'] + 1):
                    region_hits[i] += 1
            start = -1
            stop = -1
            for i, hit_count in enumerate(region_hits):  # detect regions of overlapping hits to combine overlapping hits
                if(hit_count == 0):
                    if(start != -1):  # new stop
                        stop = i - 1
                        if(ori_type == bc.FEATURE_ORIC and seq['type'] == bc.REPLICON_PLASMID):
                            ori_type = bc.FEATURE_ORIV
                        ori = OrderedDict()
                        ori['type'] = ori_type
                        ori['sequence'] = seq['id']
                        ori['start'] = start
                        ori['stop'] = stop
                        ori['strand'] = bc.STRAND_UNKNOWN
                        oris.append(ori)
                        (refined_start, refined_stop) = refine_ori_region(region_hits, ori)
                        ori['start'] = refined_start
                        ori['stop'] = refined_stop

                        if(ori_type == bc.FEATURE_ORIT):
                            ori['product'] = 'origin of transfer'
                        else:
                            ori['product'] = 'origin of replication'

                        nt = bu.extract_feature_sequence(ori, seq)  # extract nt sequences
                        ori['nt'] = nt

                        log.info(
                            'type=%s, seq=%s, start=%i, stop=%i, nt=[%s..%s]',
                            ori_type, ori['sequence'], ori['start'], ori['stop'], nt[:10], nt[-10:]
                        )
                        start = -1
                        stop = -1
                else:
                    if(start == -1):  # new start
                        start = i
    log.info('predicted=%i', len(oris))
    return oris


def refine_ori_region(region_hits: Sequence[int], ori: dict):
    log.debug('refine ori: [%i-%i]', ori['start'], ori['stop'])
    hit_region = region_hits[ori['start']:ori['stop'] + 1]
    hit_min = min(hit_region)
    hit_max = max(hit_region)
    hit_mean = sum(hit_region) / len(hit_region)
    log.debug('ori hit peak height: min=%i, max=%i, mean=%f', hit_min, hit_max, hit_mean)
    start = -1
    stop = -1
    for i in range(ori['start'], ori['stop'] + 2):  # extend range by extra bp to detect stop at 3' flank
        hit_value = region_hits[i]
        if(hit_value < hit_max / 2):
            if(start != -1):  # new stop
                stop = i - 1
                log.debug('new stop=%i', stop)
                return (start, stop)
        else:
            if(start == -1):  # new start
                start = i
                log.debug('new start=%i', start)
    return (start, ori['stop'])



import json
import subprocess as sp

from pathlib import Path
from subprocess import run

import pytest

import bakta.constants as bc

from .conftest import FILES, FILES_IO, SKIP_PARAMETERS


@pytest.mark.parametrize(
    'db',
    [
        ('db'),  # full DB
        ('db-light')  # light DB
    ]
)
def test_bakta_mock_skipped_features(db, tmpdir):
    # fast test full features
    proc = run(
        ['bin/bakta', '--db', f'test/{db}', '--output', tmpdir, '--force', '--prefix', 'test', '--min-contig-length', '200', '--proteins', 'test/data/user-proteins.faa'] +
        ['--genus', 'Foo gen. nov.', '--species', 'bar sp. nov.', '--strain', 'test 1'] +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0

    tmpdir_path = Path(tmpdir)
    for file in FILES:
        assert Path.exists(tmpdir_path.joinpath(file))


def test_bakta_plasmid(tmpdir):
    # full test on plasmid
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--verbose', '--output', tmpdir, '--force', '--prefix', 'test', '--min-contig-length', '200', '--complete', '--gram', '-', '--proteins', 'test/data/user-proteins.faa'] +
        ['--genus', 'Foo gen. nov.', '--species', 'bar sp. nov.', '--strain', 'test 1', 'test/data/NC_002127.1.fna.gz']
    )
    assert proc.returncode == 0

    tmpdir_path = Path(tmpdir)
    for file in FILES:
        output_path = tmpdir_path.joinpath(file)
        assert Path.exists(output_path)
        assert output_path.stat().st_size > 0

    results_path = tmpdir_path.joinpath('test.json')
    results = None
    with results_path.open() as fh:
        results = json.load(fh)
    assert results is not None
    features = results['features']
    assert len(features) == 3
    feature_counts_expected = {
        bc.FEATURE_T_RNA: 0,
        bc.FEATURE_TM_RNA: 0,
        bc.FEATURE_R_RNA: 0,
        bc.FEATURE_NC_RNA: 0,
        bc.FEATURE_NC_RNA_REGION: 0,
        bc.FEATURE_CRISPR: 0,
        bc.FEATURE_CDS: 3,
        bc.FEATURE_SORF: 0,
        bc.FEATURE_ORIC: 0,
        bc.FEATURE_ORIV: 0,
        bc.FEATURE_ORIT: 0
    }
    for type, count in feature_counts_expected.items():
        assert len([feat for feat in features if feat['type'] == type]) == count


@pytest.mark.parametrize(
    'db',
    [
        pytest.param('db', marks=pytest.mark.dependency(name="test_bakta_genome")),  # full DB
        ('db-light')  # light DB
    ]
)
def test_bakta_genome(db, tmpdir, request):
    # full test on complete genome in compliant mode
    output_path = Path(tmpdir).joinpath(db)
    proc = run(
        [
            'bin/bakta',
            '--db', f'test/{db}',
            '--verbose',
            '--output', str(output_path),
            '--force',
            '--prefix', 'test',
            '--min-contig-length', '200',
            '--complete',
            '--genus', 'Foo gen. nov.',
            '--species', 'bar sp. nov.',
            '--strain', 'test 1',
            '--replicons', 'test/data/replicons.tsv',
            '--regions', 'test/data/NC_002127.1-region.gff3',
            '--proteins', 'test/data/user-proteins.faa',
            '--hmms', 'test/data/NF000185.2.HMM',
            '--compliant',
            'test/data/GCF_000008865.2.fna.gz']
    )
    assert proc.returncode == 0

    for file in FILES:
        file_path = output_path.joinpath(file)
        assert Path.exists(file_path)
        assert file_path.stat().st_size > 0

    results_path = output_path.joinpath('test.json')
    if(db == 'db'):
        request.config.cache.set('test_bakta_json', str(results_path))  # cache for later re-usage

    results = None
    with results_path.open() as fh:
        results = json.load(fh)
    assert results is not None
    features = results['features']
    assert len(features) == 5550
    feature_counts_expected = {
        bc.FEATURE_T_RNA: 107,
        bc.FEATURE_TM_RNA: 1,
        bc.FEATURE_R_RNA: 7,
        bc.FEATURE_NC_RNA: 57,
        bc.FEATURE_NC_RNA_REGION: 1,
        bc.FEATURE_CRISPR: 1,
        bc.FEATURE_CDS: 5374,
        bc.FEATURE_SORF: 2,
        bc.FEATURE_ORIC: 0,
        bc.FEATURE_ORIV: 0,
        bc.FEATURE_ORIT: 0
    }
    for type, count in feature_counts_expected.items():
        assert len([feat for feat in features if feat['type'] == type]) == count


@pytest.mark.dependency(depends=['test_bakta_genome'])
def test_bakta_io(tmpdir, request):
    test_bakta_json = request.config.cache.get('test_bakta_json', None)
    assert test_bakta_json is not None
    result_path = Path(test_bakta_json)
    assert Path.exists(result_path)

    output_path = Path(tmpdir)
    proc = run(
        [
            'bin/bakta_io',
            '--verbose',
            '--output', str(output_path),
            '--force',
            '--prefix', 'test',
            test_bakta_json
        ]
    )
    assert proc.returncode == 0

    for file in FILES_IO:
        file_path = output_path.joinpath(file)
        assert Path.exists(file_path)
        assert file_path.stat().st_size > 0



import json

from pathlib import Path
from subprocess import run

import bakta.config as cfg
import bakta.expert.protein_sequences as exp_aa_seq

import pytest


SEQUENCE = 'MRADEEPGDLSAVAQDYLKVIWTAQEWSQDKVSTKMLAERIGVSASTASESIRKLAEQGLVDHEKYGAVTLTDSGRRAALAMVRRHRLLETFLVNELGYRWDEVHDEA'


AA_MIN = {
    'id': 'min',
    'description': '~~~product~~~',
    'sequence': SEQUENCE
}
AA_MIN_GENE = {
    'id': 'min',
    'description': 'gene~~~product~~~',
    'sequence': SEQUENCE
}
AA_MIN_DBXREF = {
    'id': 'min',
    'description': 'gene~~~product~~~db-1:id-1',
    'sequence': SEQUENCE
}
AA_MIN_DBXREFS = {
    'id': 'min',
    'description': 'gene~~~product~~~db-1:id-1,db-2:id-2',
    'sequence': SEQUENCE
}
AA_FULL = {
    'id': 'full',
    'description': '90.0~~~80.0~~~80.0~~~gene~~~product~~~db-1:id-1,db-2:id-2',
    'sequence': SEQUENCE
}


@pytest.mark.parametrize(
    "parameters",
    [
        AA_MIN,
        AA_MIN_GENE,
        AA_MIN_DBXREF,
        AA_MIN_DBXREFS,
        AA_FULL
    ]
)
def test_user_proteins_io(parameters, tmpdir):
    tmpdir = Path(tmpdir)
    cfg.user_proteins = tmpdir.joinpath('user.faa')
    write_tmp_faa(parameters, cfg.user_proteins)

    user_proteins_path = tmpdir.joinpath('user-clean.faa')
    exp_aa_seq.write_user_protein_sequences(user_proteins_path)


def write_tmp_faa(aa, aa_path):
    with aa_path.open('w') as fh:
        fh.write(f">{aa['id']} {aa['description']}\n{aa['sequence']}\n")


@pytest.mark.parametrize(
    "parameters",
    [
        'test/data/user-proteins.faa',
        'test/data/user-proteins.faa.gz',
        'test/data/user-proteins.gbff',
        'test/data/user-proteins.gbff.gz'
    ]
)
def test_user_proteins(parameters, tmpdir):
    # fast test skipping all feature detections
    proc = run(
        [
            'bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--prefix', 'test', '--proteins', parameters,
            '--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-crispr', '--skip-sorf', '--skip-ori', '--skip-gap', '--skip-plot',
            'test/data/NC_002127.1.fna'
        ]
    )
    assert proc.returncode == 0

    tmpdir_path = Path(tmpdir)
    results_path = Path(tmpdir_path.joinpath('test.json'))
    assert Path.exists(results_path)
    results = None
    with results_path.open() as fh:
        results = json.load(fh)
    assert results is not None
    
    ec_annotated = 0
    user_prot_feats = []
    for feat in results['features']:
        for expert in feat.get('expert', []):
            if(expert['type'] == 'user_proteins'):
                user_prot_feats.append(feat)
                ec_annotated += len([x for x in feat['db_xrefs'] if 'EC' in x])
    assert ec_annotated == 1
    assert len(user_prot_feats) == 1



import concurrent.futures as cf
import logging
import math
import subprocess as sp

from collections import OrderedDict
from typing import Sequence

from Bio.Seq import Seq

import bakta.config as cfg
import bakta.constants as bc
import bakta.features.orf as orf
import bakta.utils as bu
import bakta.psc as psc
import bakta.pscc as pscc
import bakta.so as so


log = logging.getLogger('S_ORF')


def extract(data: dict):
    """Predict open reading frames in mem via BioPython."""
    orfs = []
    for seq in data['sequences']:
        nt_seq = Seq(seq['nt'])
        for strand, strand_nt_seq in [(bc.STRAND_FORWARD, nt_seq), (bc.STRAND_REVERSE, nt_seq.reverse_complement())]:  # strands +/-
            for frame in range(3):  # frames 1/2/3 -> 0, 1, 2
                seq_frame = strand_nt_seq[frame:]

                # remove non-triplet tail nucleotides
                residue = len(seq_frame) % 3
                if(residue != 0):
                    seq_frame = seq_frame[:-residue]

                aa_seq = str(seq_frame.translate(table=cfg.translation_table, stop_symbol='*', to_stop=False, cds=False))
                aa_start = aa_seq.find('M')
                aa_end = aa_seq.find('*', aa_start)
                while aa_start > -1 and aa_end > -1:
                    orf_length = aa_end - aa_start
                    if(orf_length >= bc.MIN_SORF_LENGTH and orf_length < bc.MAX_SORF_LENGTH):  # get all CDS starts (M)
                        aa = aa_seq[aa_start:aa_end]
                        (aa_digest, aa_hexdigest) = bu.calc_aa_hash(aa)
                        if(strand == bc.STRAND_FORWARD):
                            dna_start = aa_start * 3 + frame + 1  # +1: 0 based idx to 1 based
                            dna_stop = aa_end * 3 + 2 + frame + 1
                        else:
                            dna_start = len(strand_nt_seq) - frame - (aa_end + 1) * 3 + 1
                            dna_stop = len(strand_nt_seq) - frame - aa_start * 3
                        
                        sorf = OrderedDict()
                        sorf['type'] = bc.FEATURE_SORF
                        sorf['sequence'] = seq['id']
                        sorf['start'] = dna_start
                        sorf['stop'] = dna_stop
                        sorf['strand'] = strand
                        sorf['gene'] = None
                        sorf['product'] = None
                        sorf['frame'] = frame + 1
                        sorf['db_xrefs'] = [so.SO_SORF.id]
                        sorf['aa'] = aa
                        sorf['aa_digest'] = aa_digest
                        sorf['aa_hexdigest'] = aa_hexdigest
                        
                        nt = bu.extract_feature_sequence(sorf, seq)  # extract nt sequences
                        sorf['nt'] = nt

                        orfs.append(sorf)
                        log.debug(
                            'seq=%s, start=%i, stop=%i, strand=%s, frame=%i, aa-length=%i, aa=%s, nt=[%s..%s]',
                            seq['id'], sorf['start'], sorf['stop'], strand, frame, len(aa), aa, nt[:10], nt[-10:]
                        )
                    aa_start = aa_seq.find('M', aa_start + 1)
                    if(aa_start > aa_end):
                        aa_end = aa_seq.find('*', aa_start)

    log.info('predicted=%i', len(orfs))
    return orfs


def get_feature_start(feature: dict) -> int:
    return feature['start'] if feature['strand'] == bc.STRAND_FORWARD else feature['stop']


def get_feature_stop(feature: dict) -> int:
    return feature['stop'] if feature['strand'] == bc.STRAND_FORWARD else feature['start']


def overlap_filter(data: dict, orfs_raw: Sequence[dict]):
    """Filter in-mem ORFs by overlapping CDSs."""
    t_rnas_per_sequence = {seq['id']: [] for seq in data['sequences']}
    for t_rna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_T_RNA]:
        t_rnas = t_rnas_per_sequence[t_rna['sequence']]
        t_rnas.append(t_rna)
    for tm_rna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_TM_RNA]:
        t_rnas = t_rnas_per_sequence[tm_rna['sequence']]
        t_rnas.append(tm_rna)

    r_rna_per_sequence = {seq['id']: [] for seq in data['sequences']}
    for r_rna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_R_RNA]:
        r_rnas = r_rna_per_sequence[r_rna['sequence']]
        r_rnas.append(r_rna)

    # nc_rnas_per_sequence = {seq['id']: [] for seq in data['sequences']}
    # for nc_rna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_NC_RNA]:
    #     nc_rnas = nc_rnas_per_sequence[nc_rna['sequence']]
    #     nc_rnas.append(nc_rna)
    # for nc_rna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_NC_RNA_REGION]:
    #     nc_rnas = nc_rnas_per_sequence[nc_rna['sequence']]
    #     nc_rnas.append(nc_rna)

    crispr_arrays_per_sequence = {seq['id']: [] for seq in data['sequences']}
    for crispr_array in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_CRISPR]:
        crispr_arrays = crispr_arrays_per_sequence[crispr_array['sequence']]
        crispr_arrays.append(crispr_array)

    cdss_per_sequence = {seq['id']: [] for seq in data['sequences']}
    for cds in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_CDS]:
        cdss = cdss_per_sequence[cds['sequence']]
        cdss.append(cds)

    sorfs_per_sequence = {seq['id']: [] for seq in data['sequences']}
    for sorf in orfs_raw:
        orfs = sorfs_per_sequence[sorf['sequence']]
        orfs.append(sorf)

    discarded_sorf_keys = set()
    with cf.ProcessPoolExecutor(max_workers=cfg.threads) as tpe:
        futures = []
        for seq in data['sequences']:
            sequence_sorfs = sorfs_per_sequence[seq['id']]
            log.debug('filter: seq=%s, # sORFs=%i', seq['id'], len(sequence_sorfs))
            if(len(sequence_sorfs) < 100):  # execute sORF filter task
                sorf_keys = filter_sorf(sequence_sorfs, cdss_per_sequence[seq['id']], r_rna_per_sequence[seq['id']], t_rnas_per_sequence[seq['id']], crispr_arrays_per_sequence[seq['id']])
                for sorf_key in [sk for sk in sorf_keys if sk is not None]:
                    discarded_sorf_keys.add(sorf_key)
            elif(len(sequence_sorfs) < 1000):  # submit sORF filter task to thread pool
                futures.append(tpe.submit(filter_sorf, sequence_sorfs, cdss_per_sequence[seq['id']], r_rna_per_sequence[seq['id']], t_rnas_per_sequence[seq['id']], crispr_arrays_per_sequence[seq['id']]))
            else:  # submit sORF chunk filter tasks to thread pool
                chunk_size = math.ceil(len(sequence_sorfs) / cfg.threads) if (len(sequence_sorfs) >= cfg.threads * 1000) else 1000
                log.debug('filter: chunk-size=%i', chunk_size)
                for i in range(0, len(sequence_sorfs), chunk_size):
                    sorf_chunk = sequence_sorfs[i:i + chunk_size]
                    log.debug('filter chunk: i=%i, chunk-elements=%i', i, len(sorf_chunk))
                    futures.append(tpe.submit(filter_sorf, sorf_chunk, cdss_per_sequence[seq['id']], r_rna_per_sequence[seq['id']], t_rnas_per_sequence[seq['id']], crispr_arrays_per_sequence[seq['id']]))
        for f in futures:
            for sorf_key in [sk for sk in f.result() if sk is not None]:
                discarded_sorf_keys.add(sorf_key)

    valid_sorfs = []
    discarded_sorfs = []
    for sorfs in sorfs_per_sequence.values():
        for sorf in sorfs:
            key = orf.get_orf_key(sorf)
            if(key in discarded_sorf_keys):
                discarded_sorfs.append(sorf)
            else:
                valid_sorfs.append(sorf)

    log.info('sORF filter: # valid=%i, # discarded=%i', len(valid_sorfs), len(discarded_sorfs))
    return valid_sorfs, discarded_sorfs


def filter_sorf(sorf_chunk: Sequence[dict], sequence_cdss: Sequence[dict], sequence_r_rnas: Sequence[dict], sequence_t_rnas: Sequence[dict], sequence_crispr_arrays: Sequence[dict]):
    discarded_sorf_keys = []
    for sorf in sorf_chunk:
        break_flag = False
        # filter CDS overlapping ORFs
        for cds in sequence_cdss:
            # log.debug('filter short ORFs by CDS: %s%i[%i->%i]', cds['strand'], cds['frame'], cds['start'], cds['stop'])
            if(sorf['strand'] == cds['strand']):
                if(sorf['frame'] == cds['frame']):
                    # fast/simple overlap detection for in frame orientation
                    if(cds['stop'] < sorf['start'] or cds['start'] > sorf['stop']):
                        continue
                    else:
                        discarded_sorf_keys.append(orf.get_orf_key(sorf))
                        break_flag = True
                        break
                else:
                    # if(sorf['start'] < cds['start'] and sorf['stop'] > cds['start']):  # ToDo: add max overlap threshold
                    #     # out-of-frame sorf partially overlapping CDS upstream
                    #     continue
                    # elif(sorf['start'] >= cds['start'] and sorf['stop'] <= cds['stop']):
                    if(sorf['start'] >= cds['start'] and sorf['stop'] <= cds['stop']):
                        # out-of-frame sorf completely overlapped by CDS
                        discarded_sorf_keys.append(orf.get_orf_key(sorf))
                        break_flag = True
                        break
                    # elif(sorf['start'] < cds['stop'] and sorf['stop'] > cds['stop']):  # ToDo: add max overlap threshold
                    #     # out-of-frame sorf partially overlapping CDS downstream
                    #     continue
            else:
                if(sorf['frame'] == cds['frame']):
                    # fast/simple overlap detection for in frame orientation
                    if(cds['stop'] < sorf['start'] or cds['start'] > sorf['stop']):
                        continue
                    else:
                        discarded_sorf_keys.append(orf.get_orf_key(sorf))
                        break_flag = True
                        break
                else:
                    # ToDo: add out-of-frame filters
                    # if(sorf['start'] < cds['start'] and sorf['stop'] > cds['start']):  # ToDo: add max overlap threshold
                    #     # out-frame sorf partially overlapping CDS upstream
                    #     continue
                    # elif(sorf['start'] >= cds['start'] and sorf['stop'] <= cds['stop']):
                    if(sorf['start'] >= cds['start'] and sorf['stop'] <= cds['stop']):
                        # out-frame sorf completely overlapped by CDS
                        discarded_sorf_keys.append(orf.get_orf_key(sorf))
                        break_flag = True
                        break
                    # elif(sorf['start'] < cds['stop'] and sorf['stop'] > cds['stop']):  # ToDo: add max overlap threshold
                    #     # out-frame sorf partially overlapping CDS downstream
                    #     continue
        if(break_flag):
            continue

        # filter rRNA overlapping ORFs
        for r_rna in sequence_r_rnas:
            # log.debug('filter short ORFs by rRNA: %s[%i->%i]', r_rna['strand'], r_rna['start'], r_rna['stop'])
            # fast/simple overlap detection for rRNAs
            if(sorf['stop'] < r_rna['start'] or sorf['start'] > r_rna['stop']):
                continue
            else:
                discarded_sorf_keys.append(orf.get_orf_key(sorf))
                break_flag = True
                break
        if(break_flag):
            continue

        # filter tRNA overlapping ORFs
        # log.debug('filter short ORFs by tRNA: %s[%i->%i]', t_rna['strand'], t_rna['start'], t_rna['stop'])
        for t_rna in sequence_t_rnas:
            # fast/simple overlap detection for tRNAs
            if(sorf['stop'] < t_rna['start'] or sorf['start'] > t_rna['stop']):
                continue
            else:
                discarded_sorf_keys.append(orf.get_orf_key(sorf))
                break_flag = True
                break
        if(break_flag):
            continue

        # filter CRISPR array overlapping ORFs
        # log.debug('filter short ORFs by CRISPR: [%i->%i]', crispr['start'], crispr['stop'])
        for crispr in sequence_crispr_arrays:
            # fast/simple overlap detection for CRISPR
            if(sorf['stop'] < crispr['start'] or sorf['start'] > crispr['stop']):
                continue
            else:
                discarded_sorf_keys.append(orf.get_orf_key(sorf))
                break_flag = True
                break
        if(break_flag):
            continue

    return discarded_sorf_keys


def annotation_filter(sorfs: Sequence[dict]) -> Sequence[dict]:
    """Filter sORFs according to available annotations."""
    valid_sorfs = []
    for sorf in sorfs:
        gene = None
        product = None

        ips = sorf.get('ips', None)
        if(ips is not None):
            tmp = ips.get('gene', '')
            if(tmp != ''):
                gene = tmp
            tmp = ips.get('product', '')
            if(tmp != ''):
                product = tmp

        psc = sorf.get('psc', None)
        if(psc is not None):
            tmp = psc.get('gene', '')
            if(tmp != ''):
                gene = tmp
            tmp = psc.get('product', '')
            if(tmp != ''):
                product = tmp

        pscc = sorf.get('pscc', None)
        if(pscc is not None):
            tmp = pscc.get('product', '')
            if(tmp != ''):
                product = tmp

        if(gene is None and product is None):
            sorf['hypothetical'] = True
        else:
            valid_sorfs.append(sorf)

    return valid_sorfs


def search_pscs(sorfs: Sequence[dict]):
    """Conduct homology search of sORFs against the full db PSCs"""
    return search(sorfs, 'full')


def search_psccs(sorfs: Sequence[dict]):
    """Conduct homology search of sORFs against the light db PSCCs"""
    return search(sorfs, 'light')


def search(sorfs: Sequence[dict], cluster_type: str):
    """Conduct homology search of sORFs against sORF db."""
    sorf_aa_path = cfg.tmp_path.joinpath('sorf.faa')
    orf.write_internal_faa(sorfs, sorf_aa_path)
    diamond_output_path = cfg.tmp_path.joinpath('diamond.sorf.tsv')
    diamond_db_path = cfg.db_path.joinpath('sorf.dmnd')
    cmd = [
        'diamond', 'blastp',
        '--db', str(diamond_db_path),
        '--query', str(sorf_aa_path),
        '--out', str(diamond_output_path),
        '--id', str(int(bc.MIN_SORF_IDENTITY * 100)),  # '90',
        '--query-cover', str(int(bc.MIN_SORF_COVERAGE * 100)),  # '90'
        '--subject-cover', str(int(bc.MIN_SORF_COVERAGE * 100)),  # '90'
        '--max-target-seqs', '1',  # single best output
        '--matrix', 'BLOSUM62',
        '--masking', '0',
        '--soft-masking', '0',
        '--motif-masking', '0',
        '--comp-based-stats', '0',
        '--evalue', '1',
        '--outfmt', '6', 'qseqid', 'sseqid', 'qlen', 'slen', 'length', 'pident', 'evalue', 'bitscore',
        '--threads', str(cfg.threads),
        '--load-threads', str(min(cfg.threads, 8)),  # limit load threads to max 8
        '--tmpdir', str(cfg.tmp_path),  # use tmp folder
        '--block-size', '3',  # slightly increase block size for faster executions
        '--ultra-sensitive',
        '--shape-mask', '11111',
        '--id2', '1',
        '--ungapped-evalue', '1',
        '--ungapped-evalue-short', '1',
        '--gapped-filter-evalue', '1',
    ]
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('sORF failed! diamond-error-code=%d', proc.returncode)
        raise Exception(f'diamond error! error code: {proc.returncode}')

    sorf_by_aa_digest = orf.get_orf_dictionary(sorfs)
    with diamond_output_path.open() as fh:
        for line in fh:
            (sorf_hash, cluster_id, query_length, subject_length, alignment_length, identity, evalue, bitscore) = line.split('\t')
            sorf = sorf_by_aa_digest[sorf_hash]
            query_cov = int(alignment_length) / len(sorf['aa'])
            subject_cov = int(alignment_length) / int(subject_length)
            identity = float(identity) / 100
            bitscore = float(bitscore)
            evalue = float(evalue)
            if(query_cov >= bc.MIN_SORF_COVERAGE and subject_cov >= bc.MIN_SORF_COVERAGE and identity >= bc.MIN_SORF_IDENTITY):
                result = {
                    'query_cov': query_cov,
                    'subject_cov': subject_cov,
                    'identity': identity,
                    'score': bitscore,
                    'evalue': evalue
                }
                result[psc.DB_PSC_COL_UNIREF90 if cluster_type == 'full' else pscc.DB_PSCC_COL_UNIREF50] = cluster_id
                sorf['psc' if cluster_type == 'full' else 'pscc'] = result
                log.info(
                    'homology: seq=%s, start=%i, stop=%i, strand=%s, aa-length=%i, query-cov=%0.3f, subject-cov=%0.3f, identity=%0.3f, score=%0.1f, evalue=%1.1e, UniRef=%s',
                    sorf['sequence'], sorf['start'], sorf['stop'], sorf['strand'], len(sorf['aa']), query_cov, subject_cov, identity, bitscore, evalue, cluster_id
                )

    sorfs_found = []
    sorfs_not_found = []
    for sorf in sorfs:
        if('psc' in sorf  or  'pscc' in sorf):
            sorfs_found.append(sorf)
        else:
            sorfs_not_found.append(sorf)
    log.info('found=%i', len(sorfs_found))
    return sorfs_found, sorfs_not_found



import atexit
import json
import logging
import os
import sys

from datetime import datetime
from pathlib import Path

from xopen import xopen

import bakta
import bakta.constants as bc
import bakta.config as cfg
import bakta.utils as bu
import bakta.io.fasta as fasta
import bakta.io.tsv as tsv
import bakta.io.gff as gff
import bakta.io.insdc as insdc
import bakta.plot as plot


log = logging.getLogger('IO')


def main():
    # parse options and arguments
    parser = bu.init_parser(sub_command='_io')
    parser.add_argument('input', metavar='<input>', help='Bakta annotations in JSON format')
    
    arg_group_io = parser.add_argument_group('Input / Output')
    arg_group_io.add_argument('--output', '-o', action='store', default=os.getcwd(), help='Output directory (default = current working directory)')
    arg_group_io.add_argument('--prefix', '-p', action='store', default=None, help='Prefix for output files')
    arg_group_io.add_argument('--force', '-f', action='store_true', help='Force overwriting existing output folder')
    
    arg_group_general = parser.add_argument_group('General')
    arg_group_general.add_argument('--help', '-h', action='help', help='Show this help message and exit')
    arg_group_general.add_argument('--verbose', '-v', action='store_true', help='Print verbose information')
    arg_group_general.add_argument('--debug', action='store_true', help='Run Bakta in debug mode. Temp data will not be removed.')
    arg_group_general.add_argument('--tmp-dir', action='store', default=None, dest='tmp_dir', help='Location for temporary files (default = system dependent auto detection)')
    arg_group_general.add_argument('--version', '-V', action='version', version=f'%(prog)s {cfg.version}')
    args = parser.parse_args()

    ############################################################################
    # Setup logging
    ############################################################################
    cfg.prefix = args.prefix if args.prefix else Path(args.input).stem
    output_path = cfg.check_output_path(args.output, args.force)
    cfg.force = args.force
    log.info('force=%s', args.force)
    
    bu.setup_logger(output_path, f'{cfg.prefix}.reconstruction', args)
    log.info('prefix=%s', cfg.prefix)
    log.info('output=%s', output_path)

    ############################################################################
    # Checks and configurations
    # - check parameters and setup global configuration
    # - test database
    # - test binary dependencies
    ############################################################################
    try:
        if args.input == '':
            raise ValueError('File path argument must be non-empty')
        annotation_path = Path(args.input).resolve()
        cfg.check_readability('annotation', annotation_path)
        cfg.check_content_size('annotation', annotation_path)
    except:
        log.error('provided annotation file not valid! path=%s', args.input)
        sys.exit(f'ERROR: annotation file ({args.input}) not valid!')
    log.info('input-path=%s', annotation_path)
    
    cfg.check_tmp_path(args)
    cfg.debug = args.debug
    log.info('debug=%s', cfg.debug)
    cfg.verbose = True if cfg.debug else args.verbose
    log.info('verbose=%s', cfg.verbose)
    
    if(cfg.verbose):
        print(f'Bakta v{cfg.version}')
        print('Options and arguments:')
        print(f'\tinput: {annotation_path}')
        print(f'\toutput: {cfg.output_path}')
        print(f'\tprefix: {cfg.prefix}')
        if(cfg.force): print(f'\tforce: {cfg.force}')
    
    if(cfg.debug):
        print(f"\nBakta runs in DEBUG mode! Temporary data will not be destroyed at: {cfg.tmp_path}")
    else:
        atexit.register(bu.cleanup, log, cfg.tmp_path)  # register cleanup exit hook
    
    ############################################################################
    # Import annotations from JSON
    ############################################################################
    print('Parse genome annotations...')
    with xopen(str(annotation_path), threads=0) as fh:
        data = json.load(fh)
    features = data['features']
    features_by_sequence = {seq['id']: [] for seq in data['sequences']}
    for feature in data['features']:
        seq_id = feature['sequence'] if 'sequence' in feature else feature['contig']  # <1.10.0 compatibility
        sequence_features = features_by_sequence.get(seq_id)
        sequence_features.append(feature)
    
    # set global config objects based on information from imported JSON document
    cfg.version = data['version']['bakta']
    cfg.db_info = {
        'type': data['version']['db']['type'],
        'major': data['version']['db']['version'].split('.')[0],
        'minor': data['version']['db']['version'].split('.')[1]
    }
    cfg.translation_table = data['genome']['translation_table']
    cfg.run_start = datetime.strptime(data['run']['start'], '%Y-%m-%d %H:%M:%S')
    cfg.run_end = datetime.strptime(data['run']['end'], '%Y-%m-%d %H:%M:%S')

    ############################################################################
    # Write output files
    # - write optional output files in GFF3/GenBank/EMBL formats
    # - measure runtime
    # - write comprehensive annotation results as JSON
    # - remove temp directory
    ############################################################################
    print(f'\nExport annotation results to: {cfg.output_path}')
    print('\thuman readable TSV...')
    tsv_path = cfg.output_path.joinpath(f'{cfg.prefix}.tsv')
    tsv.write_features(data['sequences'], features_by_sequence, tsv_path)

    print('\tGFF3...')
    gff3_path = cfg.output_path.joinpath(f'{cfg.prefix}.gff3')
    gff.write_features(data, features_by_sequence, gff3_path)

    print('\tINSDC GenBank & EMBL...')
    genbank_path = cfg.output_path.joinpath(f'{cfg.prefix}.gbff')
    embl_path = cfg.output_path.joinpath(f'{cfg.prefix}.embl')
    insdc.write_features(data, features, genbank_path, embl_path)

    print('\tgenome sequences...')
    fna_path = cfg.output_path.joinpath(f'{cfg.prefix}.fna')
    fasta.export_sequences(data['sequences'], fna_path, description=True, wrap=True)

    print('\tfeature nucleotide sequences...')
    ffn_path = cfg.output_path.joinpath(f'{cfg.prefix}.ffn')
    fasta.write_ffn(features, ffn_path)

    print('\ttranslated CDS sequences...')
    faa_path = cfg.output_path.joinpath(f'{cfg.prefix}.faa')
    fasta.write_faa(features, faa_path)

    print('\tfeature inferences...')
    tsv_path = cfg.output_path.joinpath(f'{cfg.prefix}.inference.tsv')
    tsv.write_feature_inferences(data['sequences'], features_by_sequence, tsv_path)

    print('\tcircular genome plot...')
    plot.write(data, features, cfg.output_path)

    hypotheticals = [feat for feat in features if feat['type'] == bc.FEATURE_CDS and 'hypothetical' in feat]
    print('\thypothetical TSV...')
    tsv_path = cfg.output_path.joinpath(f'{cfg.prefix}.hypotheticals.tsv')
    tsv.write_hypotheticals(hypotheticals, tsv_path)
    
    print('\ttranslated hypothetical CDS sequences...')
    faa_path = cfg.output_path.joinpath(f'{cfg.prefix}.hypotheticals.faa')
    fasta.write_faa(hypotheticals, faa_path)

    print('\tGenome and annotation summary...')
    summary_path = cfg.output_path.joinpath(f'{cfg.prefix}.txt')
    with summary_path.open('w') as fh_out:
        fh_out.write('Sequence(s):\n')
        fh_out.write(f"Length: {data['stats']['size']:}\n")
        fh_out.write(f"Count: {len(data['sequences'])}\n")
        fh_out.write(f"GC: {100 * data['stats']['gc']:.1f}\n")
        fh_out.write(f"N50: {data['stats']['n50']:}\n")
        if('n90' in data['stats']):
            fh_out.write(f"N90: {data['stats']['n90']:}\n")
        fh_out.write(f"N ratio: {100 * data['stats']['n_ratio']:.1f}\n")
        fh_out.write(f"coding density: {100 * data['stats']['coding_ratio']:.1f}\n")
        fh_out.write('\nAnnotation:\n')
        fh_out.write(f"tRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_T_RNA])}\n")
        fh_out.write(f"tmRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_TM_RNA])}\n")
        fh_out.write(f"rRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_R_RNA])}\n")
        fh_out.write(f"ncRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_NC_RNA])}\n")
        fh_out.write(f"ncRNA regions: {len([feat for feat in features if feat['type'] == bc.FEATURE_NC_RNA_REGION])}\n")
        fh_out.write(f"CRISPR arrays: {len([feat for feat in features if feat['type'] == bc.FEATURE_CRISPR])}\n")
        cdss = [feat for feat in features if feat['type'] == bc.FEATURE_CDS]
        fh_out.write(f"CDSs: {len(cdss)}\n")
        fh_out.write(f"pseudogenes: {len([cds for cds in cdss if 'pseudogene' in cds])}\n")
        fh_out.write(f"hypotheticals: {len([cds for cds in cdss if 'hypothetical' in cds])}\n")
        fh_out.write(f"sORFs: {len([feat for feat in features if feat['type'] == bc.FEATURE_SORF])}\n")
        fh_out.write(f"gaps: {len([feat for feat in features if feat['type'] == bc.FEATURE_GAP])}\n")
        fh_out.write(f"oriCs: {len([feat for feat in features if feat['type'] == bc.FEATURE_ORIC])}\n")
        fh_out.write(f"oriVs: {len([feat for feat in features if feat['type'] == bc.FEATURE_ORIV])}\n")
        fh_out.write(f"oriTs: {len([feat for feat in features if feat['type'] == bc.FEATURE_ORIT])}\n")
        fh_out.write('\nBakta:\n')
        fh_out.write(f'Software: v{cfg.version}\n')
        fh_out.write(f"Database: v{cfg.db_info['major']}.{cfg.db_info['minor']}, {cfg.db_info['type']}\n")
        fh_out.write('DOI: 10.1099/mgen.0.000685\n')
        fh_out.write('URL: github.com/oschwengers/bakta\n')


if __name__ == '__main__':
    main()



import atexit
import logging
import sys

from datetime import datetime
from pathlib import Path

import bakta
import bakta.constants as bc
import bakta.config as cfg
import bakta.io.fasta as fasta
import bakta.io.json as json
import bakta.io.tsv as tsv
import bakta.io.gff as gff
import bakta.io.insdc as insdc
import bakta.expert.amrfinder as exp_amr
import bakta.expert.protein_sequences as exp_aa_seq
import bakta.expert.protein_hmms as exp_aa_hmms
import bakta.features.annotation as anno
import bakta.features.t_rna as t_rna
import bakta.features.tm_rna as tm_rna
import bakta.features.r_rna as r_rna
import bakta.features.nc_rna as nc_rna
import bakta.features.nc_rna_region as nc_rna_region
import bakta.features.crispr as crispr
import bakta.features.orf as orf
import bakta.features.cds as feat_cds
import bakta.features.s_orf as s_orf
import bakta.features.gaps as gaps
import bakta.features.ori as ori
import bakta.db as db
import bakta.utils as bu
import bakta.ups as ups
import bakta.ips as ips
import bakta.psc as psc
import bakta.pscc as pscc
import bakta.plot as plot


def main():
    args = bu.parse_arguments()  # parse arguments

    ############################################################################
    # Setup logging
    ############################################################################
    cfg.prefix = args.prefix if args.prefix else Path(args.genome).stem
    output_path = cfg.check_output_path(args.output, args.force)
    bu.setup_logger(output_path, cfg.prefix, args)
    log = logging.getLogger('MAIN')

    ############################################################################
    # Checks and configurations
    # - check parameters and setup global configuration
    # - test database
    # - test binary dependencies
    ############################################################################
    cfg.setup(args)  # check parameters and prepare global configuration
    cfg.db_info = db.check(cfg.db_path)
    bu.test_dependencies()
    if(cfg.verbose):
        print(f'Bakta v{cfg.version}')
        print('Options and arguments:')
        print(f'\tinput: {cfg.genome_path}')
        print(f"\tdb: {cfg.db_path}, version {cfg.db_info['major']}.{cfg.db_info['minor']}, {cfg.db_info['type']}")
        if(cfg.replicons): print(f'\treplicon table: {cfg.replicons}')
        if(cfg.prodigal_tf): print(f'\tprodigal training file: {cfg.prodigal_tf}')
        if(cfg.regions): print(f'\tregion table: {cfg.regions}')
        if(cfg.user_proteins): print(f'\tuser proteins: {cfg.user_proteins}')
        if(cfg.user_hmms): print(f'\tuser hmms: {cfg.user_hmms}')
        print(f'\ttranslation table: {cfg.translation_table}')
        if(cfg.taxon): print(f'\ttaxon: {cfg.taxon}')
        if(cfg.plasmid): print(f'\tplasmid: {cfg.plasmid}')
        if(cfg.gram != '?'): print(f'\tgram: {cfg.gram}')
        if(cfg.locus): print(f'\tlocus prefix: {cfg.locus}')
        if(cfg.locus_tag): print(f'\tlocus tag prefix: {cfg.locus_tag}')
        if(cfg.meta): print(f'\tmeta mode: {cfg.meta}')
        if(cfg.partial): print(f'\tpredict partial genes: {cfg.partial}')
        if(cfg.complete): print(f'\tcomplete replicons: {cfg.complete}')
        print(f'\toutput: {cfg.output_path}')
        if(cfg.force): print(f'\tforce: {cfg.force}')
        print(f'\ttmp directory: {cfg.tmp_path}')
        if(cfg.compliant): print(f'\tINSDC compliant: {cfg.compliant}')
        if(cfg.keep_sequence_headers): print(f'\tkeep/sequence headers: {cfg.keep_sequence_headers}')
        print(f'\tprefix: {cfg.prefix}')
        print(f'\tthreads: {cfg.threads}')
        if(cfg.debug): print(f'\tdebug: {cfg.debug}')
        if(cfg.skip_trna): print(f'\tskip tRNA: {cfg.skip_trna}')
        if(cfg.skip_tmrna): print(f'\tskip tmRNA: {cfg.skip_tmrna}')
        if(cfg.skip_rrna): print(f'\tskip rRNA: {cfg.skip_rrna}')
        if(cfg.skip_ncrna): print(f'\tskip ncRNA: {cfg.skip_ncrna}')
        if(cfg.skip_ncrna_region): print(f'\tskip ncRNA region: {cfg.skip_ncrna_region}')
        if(cfg.skip_crispr): print(f'\tskip CRISPR: {cfg.skip_crispr}')
        if(cfg.skip_cds): print(f'\tskip CDS: {cfg.skip_cds}')
        if(cfg.skip_sorf): print(f'\tskip sORF: {cfg.skip_sorf}')
        if(cfg.skip_gap): print(f'\tskip gap: {cfg.skip_gap}')
        if(cfg.skip_ori): print(f'\tskip oriC/V/T: {cfg.skip_ori}')
        if(cfg.skip_filter): print(f'\tskip feature overlap filters: {cfg.skip_filter}')
        if(cfg.skip_plot): print(f'\tskip plot: {cfg.skip_plot}')
        print()
    
    if(cfg.debug):
        print(f"Bakta runs in DEBUG mode! Temporary data will not be destroyed at: {cfg.tmp_path}\n")
    else:
        atexit.register(bu.cleanup, log, cfg.tmp_path)  # register cleanup exit hook

    ############################################################################
    # Import genome
    # - parse sequences in Fasta file
    # - apply sequence length filter
    # - rename sequences
    ############################################################################
    print('Parse genome sequences...')
    try:
        sequences = fasta.import_sequences(cfg.genome_path)
        log.info('imported sequences=%i', len(sequences))
        print(f'\timported: {len(sequences)}')
    except:
        log.error('wrong genome file format!', exc_info=True)
        sys.exit('ERROR: wrong genome file format!')
    replicons = bu.parse_replicon_table(cfg.replicons) if cfg.replicons else None
    sequences, complete_genome = bu.qc_sequences(sequences, replicons)
    print(f'\tfiltered & revised: {len(sequences)}')
    no_chromosomes = len([seq for seq in sequences if seq['type'] == bc.REPLICON_CHROMOSOME])
    if(no_chromosomes > 0):
        print(f"\tchromosomes: {no_chromosomes}")
    no_plasmids = len([seq for seq in sequences if seq['type'] == bc.REPLICON_PLASMID])
    if(no_plasmids > 0):
        print(f"\tplasmids: {no_plasmids}")
    no_contigs = len([seq for seq in sequences if seq['type'] == bc.REPLICON_CONTIG])
    if(no_contigs > 0):
        print(f"\tcontigs: {no_contigs}")
    if(len(sequences) == 0):
        log.warning('no valid sequences!')
        sys.exit('Error: input file contains no valid sequences.')
    sequences_path = cfg.tmp_path.joinpath('sequences.fna')
    fasta.export_sequences(sequences, sequences_path)
    data = {
        'genome': {
            'genus': cfg.genus,
            'species': cfg.species,
            'strain': cfg.strain,
            'taxon': cfg.taxon,
            'complete': cfg.complete or complete_genome,
            'gram': cfg.gram,
            'translation_table': cfg.translation_table
        },
        'stats': {
            'size': sum([seq['length'] for seq in sequences])
        },
        'features': [],
        'sequences': sequences
    }
    if(cfg.plasmid):
        data['genome']['plasmid'] = cfg.plasmid
    print('\nStart annotation...')

    ############################################################################
    # tRNA prediction
    ############################################################################
    if(cfg.skip_trna):
        print('skip tRNA prediction...')
    else:
        print('predict tRNAs...')
        log.debug('start tRNA prediction')
        trnas = t_rna.predict_t_rnas(data, sequences_path)
        data['features'].extend(trnas)
        print(f"\tfound: {len(trnas)}")

    ############################################################################
    # tmRNA prediction
    ############################################################################
    if(cfg.skip_tmrna):
        print('skip tmRNA prediction...')
    else:
        print('predict tmRNAs...')
        log.debug('start tmRNA prediction')
        tmrnas = tm_rna.predict_tm_rnas(data, sequences_path)
        data['features'].extend(tmrnas)
        print(f"\tfound: {len(tmrnas)}")

    ############################################################################
    # rRNA prediction
    ############################################################################
    if(cfg.skip_rrna):
        print('skip rRNA prediction...')
    else:
        print('predict rRNAs...')
        log.debug('start rRNA prediction')
        rrnas = r_rna.predict_r_rnas(data, sequences_path)
        data['features'].extend(rrnas)
        print(f"\tfound: {len(rrnas)}")

    ############################################################################
    # ncRNA gene prediction
    ############################################################################
    if(cfg.skip_ncrna):
        print('skip ncRNA prediction...')
    else:
        print('predict ncRNAs...')
        log.debug('start ncRNA prediction')
        ncrnas = nc_rna.predict_nc_rnas(data, sequences_path)
        data['features'].extend(ncrnas)
        print(f"\tfound: {len(ncrnas)}")

    ############################################################################
    # ncRNA region prediction
    ############################################################################
    if(cfg.skip_ncrna_region):
        print('skip ncRNA region prediction...')
    else:
        print('predict ncRNA regions...')
        log.debug('start ncRNA region prediction')
        ncrna_regions = nc_rna_region.predict_nc_rna_regions(data, sequences_path)
        data['features'].extend(ncrna_regions)
        print(f"\tfound: {len(ncrna_regions)}")

    ############################################################################
    # CRISPR prediction
    ############################################################################
    if(cfg.skip_crispr):
        print('skip CRISPR array prediction...')
    else:
        print('predict CRISPR arrays...')
        log.debug('start CRISPR prediction')
        crisprs = crispr.predict_crispr(data, sequences_path)
        data['features'].extend(crisprs)
        print(f"\tfound: {len(crisprs)}")

    ############################################################################
    # CDS prediction
    # - Prodigal prediction
    # - lookup UPS matches
    # - lookup IPS matches
    # - search PSC for unannotated CDSs
    # - conduct expert systems analysis
    # - lookup & combine annotations
    # - analyze hypotheticals
    ############################################################################
    if(cfg.skip_cds):
        print('skip CDS prediction...')
    else:
        print('predict & annotate CDSs...')
        log.debug('predict CDS')
        cdss = feat_cds.predict(data)
        print(f"\tpredicted: {len(cdss)} ")

        if(len(cdss) > 0):
            log.debug('discard too-long CDS')
            discarded_cdss = feat_cds.filter_length(cdss)
            print(f'\tdiscarded length: {len(discarded_cdss)}')
            cdss = [cds for cds in cdss if 'discarded' not in cds]

        if(len(cdss) > 0):
            log.debug('detect spurious CDS')
            discarded_cdss = orf.detect_spurious(cdss)
            print(f'\tdiscarded spurious: {len(discarded_cdss)}')
            cdss = [cds for cds in cdss if 'discarded' not in cds]
        
        if(len(cdss) > 0):
            log.debug('revise translational exceptions')
            no_revised = feat_cds.revise_translational_exceptions(data, cdss)
            print(f'\trevised translational exceptions: {no_revised}')
            cdss = [cds for cds in cdss if 'discarded' not in cds]
        
        if(cfg.regions):
            log.debug('import user-provided CDS regions')
            imported_cdss, no_skipped = feat_cds.import_user_cdss(data, cfg.regions)
            print(f'\tuser regions: {len(imported_cdss)} imported, {no_skipped} skipped')
            cdss.extend(imported_cdss)

        if(len(cdss) > 0):
            if(cfg.db_info['type'] == 'full'):
                log.debug('lookup CDS UPS/IPS')
                cdss_ups, cdss_not_found_ups = ups.lookup(cdss)
                cdss_ips, cdss_not_found_ips = ips.lookup(cdss_ups)
                cdss_not_found = cdss_not_found_ups + cdss_not_found_ips
                print(f'\tdetected IPSs: {len(cdss_ips)}')
            else:
                cdss_not_found = [*cdss]
                print(f'\tskip UPS/IPS detection with light db version')

            if(len(cdss_not_found) > 0):
                if(cfg.db_info['type'] == 'full'):
                    log.debug('search CDS PSC')
                    cdss_psc, cdss_pscc, cdss_not_found = psc.search(cdss_not_found)
                    print(f'\tfound PSCs: {len(cdss_psc)}')
                    print(f'\tfound PSCCs: {len(cdss_pscc)}')
                else:
                    log.debug('search CDS PSCC')
                    cdss_pscc, cdss_not_found = pscc.search(cdss_not_found)
                    print(f'\tfound PSCCs: {len(cdss_pscc)}')
            print('\tlookup annotations...')
            log.debug('lookup CDS PSCs')
            psc.lookup(cdss)  # lookup PSC info
            pscc.lookup(cdss)  # lookup PSCC info

            print('\tconduct expert systems...')  # conduct expert systems annotation
            cds_aa_path = cfg.tmp_path.joinpath('cds.expert.faa')
            orf.write_internal_faa(cdss, cds_aa_path)
            log.debug('conduct expert system: amrfinder')
            expert_amr_found = exp_amr.search(cdss, cds_aa_path)
            print(f'\t\tamrfinder: {len(expert_amr_found)}')
            log.debug('conduct expert system: aa seqs')
            diamond_db_path = cfg.db_path.joinpath('expert-protein-sequences.dmnd')
            expert_aa_found = exp_aa_seq.search(cdss, cds_aa_path, 'expert_proteins', diamond_db_path)
            print(f'\t\tprotein sequences: {len(expert_aa_found)}')

            if(cfg.user_proteins):
                log.debug('conduct expert system: user aa seqs')
                user_aa_path = cfg.tmp_path.joinpath('user-proteins.faa')
                no_user_proteins, no_skipped_user_proteins = exp_aa_seq.write_user_protein_sequences(user_aa_path)
                if(no_skipped_user_proteins > 0):
                    print(f'\t\tWARNING: skipped incorrectly formatted user proteins: {no_skipped_user_proteins}, imported: {no_user_proteins}')
                user_aa_found = exp_aa_seq.search(cdss, cds_aa_path, 'user_proteins', user_aa_path)
                print(f'\t\tuser protein sequences: {len(user_aa_found)}')

            if(cfg.user_hmms):
                log.debug('conduct expert system: user HMM')
                user_hmm_found = exp_aa_hmms.search(cdss, cfg.user_hmms)
                print(f'\t\tuser HMM sequences: {len(user_hmm_found)}')

            print('\tcombine annotations and mark hypotheticals...')
            log.debug('combine CDS annotations')
            for cds in cdss:
                anno.combine_annotation(cds)  # combine IPS & PSC annotations and mark hypotheticals

            hypotheticals = [cds for cds in cdss if 'hypothetical' in cds and 'edge' not in cds and cds.get('start_type', 'Edge') != 'Edge']
            if(len(hypotheticals) > 0  and  not cfg.skip_pseudo):
                if(cfg.db_info['type'] == 'full'):
                    print('\tdetect pseudogenes...')
                    log.debug('search pseudogene candidates')
                    pseudo_candidates = feat_cds.predict_pseudo_candidates(hypotheticals)
                    print(f'\t\tcandidates: {len(pseudo_candidates)}')
                    pseudogenes = feat_cds.detect_pseudogenes(pseudo_candidates, cdss, data) if len(pseudo_candidates) > 0 else []
                    psc.lookup(pseudogenes, pseudo=True)
                    pscc.lookup(pseudogenes, pseudo=True)
                    for pseudogene in pseudogenes:
                        anno.combine_annotation(pseudogene)
                    print(f'\t\tverified: {len(pseudogenes)}')
                else:
                    print(f'\tskip pseudogene detection with light db version')
            hypotheticals = [cds for cds in cdss if 'hypothetical' in cds]
            if(len(hypotheticals) > 0):
                log.debug('analyze hypotheticals')
                print(f'\tanalyze hypothetical proteins: {len(hypotheticals)}')
                pfam_hits = feat_cds.predict_pfam(hypotheticals)
                print(f"\t\tdetected Pfam hits: {len(pfam_hits)} ")
                feat_cds.analyze_proteins(hypotheticals)
                print('\t\tcalculated proteins statistics')
            
            print('\trevise special cases...')
            feat_cds.revise_special_cases_annotated(data, cdss)

        data['features'].extend(cdss)

    ############################################################################
    # sORF prediction
    # - in-mem sORF extraction
    # - overlap filtering (tRNA, tmRNA, rRNA, CDS)
    # - lookup UPS matches
    # - lookup IPS matches
    # - filter sORFs w/o IPS match
    ############################################################################
    if(cfg.skip_sorf):
        print('skip sORF prediction...')
    else:
        print('detect & annotate sORF...')
        log.debug('extract sORF')
        sorfs = s_orf.extract(data)
        print(f'\tdetected: {len(sorfs)}')

        log.debug('apply sORF overlap filter')
        sorfs, discarded_sorfs = s_orf.overlap_filter(data, sorfs)
        print(f'\tdiscarded due to overlaps: {len(discarded_sorfs)}')

        if(len(sorfs) > 0):
            log.debug('detect spurious sORF')
            discarded_sorfs = orf.detect_spurious(sorfs)
            print(f'\tdiscarded spurious: {len(discarded_sorfs)}')
            sorfs = [sorf for sorf in sorfs if 'discarded' not in sorf]

        log.debug('lookup sORF UPS/IPS')
        sorf_upss, sorfs_not_found = ups.lookup(sorfs)
        sorf_ipss, tmp = ips.lookup(sorf_upss)
        sorfs_not_found.extend(tmp)
        print(f'\tdetected IPSs: {len(sorf_ipss)}')

        sorf_pscs_psccs = []
        if(len(sorfs_not_found) > 0):
            if(cfg.db_info['type'] == 'full'):
                log.debug('search sORF PSC')
                cdss_not_found_tmp, sorfs_not_found = s_orf.search_pscs(sorfs_not_found)
                sorf_pscs_psccs.extend(cdss_not_found_tmp)
                print(f'\tfound PSCs: {len(sorf_pscs_psccs)}')
            else:
                log.debug('search sORF PSCC')
                sorf_psccs, sorfs_not_found = s_orf.search_psccs(sorfs_not_found)
                sorf_pscs_psccs.extend(sorf_psccs)
                print(f'\tfound PSCCs: {len(sorf_pscs_psccs)}')


        print("\tlookup annotations...")
        log.debug('lookup sORF PSCs')
        sorf_pscs_psccs.extend(sorf_ipss)
        psc.lookup(sorf_pscs_psccs)  # lookup PSC info
        log.debug('lookup sORF PSCCs')
        pscc.lookup(sorf_pscs_psccs)  # lookup PSC info
        print('\tfilter and combine annotations...')
        log.debug('filter sORF by annotations')
        sorfs_filtered = s_orf.annotation_filter(sorfs)
        log.debug('combine sORF annotations')
        for feat in sorfs_filtered:
            anno.combine_annotation(feat)  # combine IPS and PSC annotations
        data['features'].extend(sorfs_filtered)
        print(f'\tfiltered sORFs: {len(sorfs_filtered)}')

    ############################################################################
    # gap annotation
    # - in-mem gap detection
    # - gap annotation
    ############################################################################
    if(cfg.skip_gap):
        print('skip gap annotation...')
    else:
        print('detect gaps...')
        log.debug('detect gaps')
        assembly_gaps = gaps.detect_assembly_gaps(data)
        data['features'].extend(assembly_gaps)
        print(f'\tfound: {len(assembly_gaps)}')

    ############################################################################
    # oriC/T prediction
    ############################################################################
    if(cfg.skip_ori):
        print('skip oriC/T annotation...')
    else:
        print('detect oriCs/oriVs...')
        log.debug('detect oriC/V')
        oriCs = ori.predict_oris(data, sequences_path, bc.FEATURE_ORIC)
        data['features'].extend(oriCs)
        print(f'\tfound: {len(oriCs)}')

        print('detect oriTs...')
        log.debug('detect oriT')
        oriTs = ori.predict_oris(data, sequences_path, bc.FEATURE_ORIT)
        data['features'].extend(oriTs)
        print(f'\tfound: {len(oriTs)}')

    ############################################################################
    # Filter overlapping features
    ############################################################################
    if(cfg.skip_filter):
        print('skip feature overlap filters...')
    else:
        print('apply feature overlap filters...')
        anno.detect_feature_overlaps(data)

    ############################################################################
    # Create annotations
    # - filter features based on precedence and overlaps
    # - sort features
    # - create locus tags for features
    ############################################################################
    print('select features and create locus tags...')
    log.debug('start feature selection and creation of locus tags')
    features_by_sequence = {seq['id']: [] for seq in data['sequences']}
    feature_id = 1
    feature_id_prefix = bu.create_locus_tag_prefix(sequences, length=10)
    for feature in data['features']:
        if('discarded' not in feature):
            feature['id'] = f'{feature_id_prefix}_{feature_id}'
            feature_id += 1
            seq_features = features_by_sequence.get(feature['sequence'])
            seq_features.append(feature)
    features = []
    for seq in data['sequences']:
        seq_features = features_by_sequence[seq['id']]
        seq_features.sort(key=lambda k: k['start'])
        features.extend(seq_features)
    data['features'] = features  # overwrite feature list by final sorted feature list
    log.info('selected features=%i', len(features))
    print(f'\tselected: {len(features)}')

    # use user provided locus tag if not None/non-empty or generate a sequence based locus prefix
    locus_tag_prefix = cfg.locus_tag if cfg.locus_tag else bu.create_locus_tag_prefix(sequences)
    log.info('locus tag prefix=%s', locus_tag_prefix)
    locus_tag_nr = cfg.locus_tag_increment
    for feature in features:
        locus_tag = f'{locus_tag_prefix}_{locus_tag_nr:0{len(str(cfg.locus_tag_increment*len(features)))+1}}'
        if(feature['type'] in [bc.FEATURE_T_RNA, bc.FEATURE_TM_RNA, bc.FEATURE_R_RNA, bc.FEATURE_NC_RNA, bc.FEATURE_CDS, bc.FEATURE_SORF]):
            feature['locus'] = locus_tag
            locus_tag_nr += cfg.locus_tag_increment

    ############################################################################
    # Improve annotations
    # - select CDS/sORF gene symbols based on adjacent genes
    ############################################################################
    print('improve annotations...')
    genes_with_improved_symbols = anno.select_gene_symbols([feature for feature in features if feature['type'] in [bc.FEATURE_CDS, bc.FEATURE_SORF]])
    print(f'\trevised gene symbols: {len(genes_with_improved_symbols)}')

    ############################################################################
    # Print summary
    # - genome stats
    # - annotation stats
    ############################################################################
    bu.calc_genome_stats(data)
    print('\nGenome statistics:')
    print(f"\tGenome size: {data['stats']['size']:,} bp")
    print(f"\tContigs/replicons: {len(data['sequences'])}")
    print(f"\tGC: {100 * data['stats']['gc']:.1f} %")
    print(f"\tN50: {data['stats']['n50']:,}")
    print(f"\tN90: {data['stats']['n90']:,}")
    print(f"\tN ratio: {100 * data['stats']['n_ratio']:.1f} %")
    print(f"\tcoding density: {100 * data['stats']['coding_ratio']:.1f} %")
    print('\nannotation summary:')
    print(f"\ttRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_T_RNA])}")
    print(f"\ttmRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_TM_RNA])}")
    print(f"\trRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_R_RNA])}")
    print(f"\tncRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_NC_RNA])}")
    print(f"\tncRNA regions: {len([feat for feat in features if feat['type'] == bc.FEATURE_NC_RNA_REGION])}")
    print(f"\tCRISPR arrays: {len([feat for feat in features if feat['type'] == bc.FEATURE_CRISPR])}")
    cdss = [feat for feat in features if feat['type'] == bc.FEATURE_CDS]
    print(f"\tCDSs: {len(cdss)}")
    print(f"\t\thypotheticals: {len([cds for cds in cdss if 'hypothetical' in cds])}")
    print(f"\t\tpseudogenes: {len([cds for cds in cdss if 'pseudogene' in cds])}")
    print(f"\tsORFs: {len([feat for feat in features if feat['type'] == bc.FEATURE_SORF])}")
    print(f"\tgaps: {len([feat for feat in features if feat['type'] == bc.FEATURE_GAP])}")
    print(f"\toriCs/oriVs: {len([feat for feat in features if (feat['type'] == bc.FEATURE_ORIC or feat['type'] == bc.FEATURE_ORIV)])}")
    print(f"\toriTs: {len([feat for feat in features if feat['type'] == bc.FEATURE_ORIT])}")

    ############################################################################
    # Write output files
    # - measure runtime
    # - write optional output files in GFF3/GenBank/EMBL formats
    # - write comprehensive annotation results as JSON
    # - remove temp directory
    ############################################################################
    cfg.run_end = datetime.now()  # measure runtime

    print(f'\nExport annotation results to: {cfg.output_path}')
    print('\thuman readable TSV...')
    tsv_path = cfg.output_path.joinpath(f'{cfg.prefix}.tsv')
    tsv.write_features(data['sequences'], features_by_sequence, tsv_path)

    print('\tGFF3...')
    gff3_path = cfg.output_path.joinpath(f'{cfg.prefix}.gff3')
    gff.write_features(data, features_by_sequence, gff3_path)

    print('\tINSDC GenBank & EMBL...')
    genbank_path = cfg.output_path.joinpath(f'{cfg.prefix}.gbff')
    embl_path = cfg.output_path.joinpath(f'{cfg.prefix}.embl')
    insdc.write_features(data, features, genbank_path, embl_path)

    print('\tgenome sequences...')
    fna_path = cfg.output_path.joinpath(f'{cfg.prefix}.fna')
    fasta.export_sequences(data['sequences'], fna_path, description=True, wrap=True)

    print('\tfeature nucleotide sequences...')
    ffn_path = cfg.output_path.joinpath(f'{cfg.prefix}.ffn')
    fasta.write_ffn(features, ffn_path)

    print('\ttranslated CDS sequences...')
    faa_path = cfg.output_path.joinpath(f'{cfg.prefix}.faa')
    fasta.write_faa(features, faa_path)

    print('\tfeature inferences...')
    tsv_path = cfg.output_path.joinpath(f'{cfg.prefix}.inference.tsv')
    tsv.write_feature_inferences(data['sequences'], features_by_sequence, tsv_path)

    if(cfg.skip_plot  or  cfg.meta):
        print('\tskip generation of circular genome plot...')
    else:
        print('\tcircular genome plot...')
        plot.write(data, features, cfg.output_path)

    if(cfg.skip_cds is False):
        hypotheticals = [feat for feat in features if feat['type'] == bc.FEATURE_CDS and 'hypothetical' in feat]
        print('\thypothetical TSV...')
        tsv_path = cfg.output_path.joinpath(f'{cfg.prefix}.hypotheticals.tsv')
        tsv.write_hypotheticals(hypotheticals, tsv_path)

        print('\ttranslated hypothetical CDS sequences...')
        faa_path = cfg.output_path.joinpath(f'{cfg.prefix}.hypotheticals.faa')
        fasta.write_faa(hypotheticals, faa_path)

    # calc & store runtime
    run_duration = (cfg.run_end - cfg.run_start).total_seconds()
    data['run'] = {
        'start': cfg.run_start.strftime('%Y-%m-%d %H:%M:%S'),
        'end': cfg.run_end.strftime('%Y-%m-%d %H:%M:%S'),
        'duration': f'{(run_duration / 60):.2f} min'
    }

    print('\tmachine readable JSON...')
    json_path = cfg.output_path.joinpath(f'{cfg.prefix}.json')
    json.write_json(data, features, json_path)

    print('\tGenome and annotation summary...')
    summary_path = cfg.output_path.joinpath(f'{cfg.prefix}.txt')
    with summary_path.open('w') as fh_out:
        fh_out.write('Sequence(s):\n')
        fh_out.write(f"Length: {data['stats']['size']:}\n")
        fh_out.write(f"Count: {len(data['sequences'])}\n")
        fh_out.write(f"GC: {100 * data['stats']['gc']:.1f}\n")
        fh_out.write(f"N50: {data['stats']['n50']:}\n")
        fh_out.write(f"N90: {data['stats']['n90']:}\n")
        fh_out.write(f"N ratio: {100 * data['stats']['n_ratio']:.1f}\n")
        fh_out.write(f"coding density: {100 * data['stats']['coding_ratio']:.1f}\n")
        fh_out.write('\nAnnotation:\n')
        fh_out.write(f"tRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_T_RNA])}\n")
        fh_out.write(f"tmRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_TM_RNA])}\n")
        fh_out.write(f"rRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_R_RNA])}\n")
        fh_out.write(f"ncRNAs: {len([feat for feat in features if feat['type'] == bc.FEATURE_NC_RNA])}\n")
        fh_out.write(f"ncRNA regions: {len([feat for feat in features if feat['type'] == bc.FEATURE_NC_RNA_REGION])}\n")
        fh_out.write(f"CRISPR arrays: {len([feat for feat in features if feat['type'] == bc.FEATURE_CRISPR])}\n")
        fh_out.write(f"CDSs: {len(cdss)}\n")
        fh_out.write(f"pseudogenes: {len([cds for cds in cdss if 'pseudogene' in cds])}\n")
        fh_out.write(f"hypotheticals: {len([cds for cds in cdss if 'hypothetical' in cds])}\n")
        fh_out.write(f"sORFs: {len([feat for feat in features if feat['type'] == bc.FEATURE_SORF])}\n")
        fh_out.write(f"gaps: {len([feat for feat in features if feat['type'] == bc.FEATURE_GAP])}\n")
        fh_out.write(f"oriCs: {len([feat for feat in features if feat['type'] == bc.FEATURE_ORIC])}\n")
        fh_out.write(f"oriVs: {len([feat for feat in features if feat['type'] == bc.FEATURE_ORIV])}\n")
        fh_out.write(f"oriTs: {len([feat for feat in features if feat['type'] == bc.FEATURE_ORIT])}\n")
        fh_out.write('\nBakta:\n')
        fh_out.write(f'Software: v{cfg.version}\n')
        fh_out.write(f"Database: v{cfg.db_info['major']}.{cfg.db_info['minor']}, {cfg.db_info['type']}\n")
        fh_out.write('DOI: 10.1099/mgen.0.000685\n')
        fh_out.write('URL: github.com/oschwengers/bakta\n')

    print(f'\nIf you use these results please cite Bakta: https://doi.org/{bc.BAKTA_DOI}')
    print(f'Annotation successfully finished in {int(run_duration / 60):01}:{int(run_duration % 60):02} [mm:ss].')


if __name__ == '__main__':
    main()



import logging
import multiprocessing as mp
import os
import re
import sys
import tempfile

from argparse import Namespace
from datetime import datetime
from pathlib import Path

import bakta
import bakta.constants as bc


PLASMID_NAME_PATTERN = re.compile(r'p[a-zA-Z0-9\._]{1,19}')
PLASMID_UNNAMED_PATTERN = re.compile(r'unnamed[0-9]{0,3}')


log = logging.getLogger('CONFIG')


# runtime configurations
env = os.environ.copy()
threads = None
verbose = None
debug = None

# input / output configuration
version = bakta.__version__
db_path = None
db_info = None
tmp_path = None
genome_path = None
min_sequence_length = None
prefix = None
output_path = None
force = None

# organism configuration
genus = None
species = None
strain = None
plasmid = None
taxon = None

# annotation configuration
complete = None
prodigal_tf = None
translation_table = None
keep_sequence_headers = None
locus = None
locus_tag = None
locus_tag_increment = None
gram = None
replicons = None
compliant = None
user_proteins = None
user_hmms = None
meta = None
partial = None
regions = None

# workflow configuration
skip_trna = None
skip_tmrna = None
skip_rrna = None
skip_ncrna = None
skip_ncrna_region = None
skip_crispr = None
skip_cds = None
skip_pseudo = None
skip_sorf = None
skip_gap = None
skip_ori = None
skip_filter = None
skip_plot = None

run_start = datetime.now()
run_end = None


def setup(args):
    """Test environment and build a runtime configuration."""
    # runtime configurations
    global env, threads, verbose, debug
    env['BLAST_USAGE_REPORT'] = 'false'  # prevent BLAST from contacting NCBI

    threads = check_threads(args)
    verbose = args.verbose
    log.info('verbose=%s', verbose)
    debug = args.debug
    log.info('debug=%s', debug)
    if(debug):
        verbose = True

    # input / output path configurations
    global db_path, db_info, tmp_path, genome_path, min_sequence_length, prefix, output_path, force
    db_path = check_db_path(args)
    tmp_path = check_tmp_path(args)

    try:
        if(args.genome == ''):
            raise ValueError('File path argument must be non-empty')
        genome_path = Path(args.genome).resolve()
        check_readability('genome', genome_path)
        check_content_size('genome', genome_path)
    except:
        log.error('provided genome file not valid! path=%s', args.genome)
        sys.exit(f'ERROR: genome file ({args.genome}) not valid!')
    log.info('genome-path=%s', genome_path)

    # input / output configurations
    min_sequence_length = args.min_contig_length
    if(min_sequence_length <= 0):
        log.error("wrong argument for 'min-contig-length' parameter! min_contig_length=%s", min_sequence_length)
        sys.exit(f"ERROR: wrong argument ({min_sequence_length}) for 'min- contig-length' parameter! Value must be larger than 0")
    log.info('min_contig_length=%s', min_sequence_length)
    log.info('prefix=%s', prefix)  # set in main.py before global logger config
    log.info('output-path=%s', output_path)
    force = args.force
    log.info('force=%s', force)

    # organism configurations
    global genus, species, strain, plasmid, taxon
    genus = args.genus
    if(genus is not None):
        genus = genus.strip()
        if(genus == ''):
            log.error("Empty 'genus' parameter! genus=%s", genus)
            sys.exit(f"ERROR: empty 'genus' parameter!")
        else:
            genus = genus.capitalize()
    log.info('genus=%s', genus)
    species = args.species
    if(species is not None):
        species = species.strip()
        if(species == ''):
            log.error("Empty 'species' parameter! species=%s", species)
            sys.exit(f"ERROR: empty 'species' parameter!")
        else:
            species = species.lower()
    log.info('species=%s', species)
    strain = args.strain
    if(strain is not None):
        strain = strain.strip()
        if(strain == ''):
            log.error("Empty 'strain' parameter! strain=%s", species)
            sys.exit(f"ERROR: empty 'strain' parameter!")
    log.info('strain=%s', strain)
    plasmid = args.plasmid
    if(plasmid is not None):
        plasmid = plasmid.strip()
        if(plasmid == ''):
            log.error("Empty 'plasmid' parameter! plasmid=%s", plasmid)
            sys.exit(f"ERROR: empty 'plasmid' parameter!")
        elif('plasmid' in plasmid.lower()):
            log.error("Wrong 'plasmid' parameter! plasmid=%s", plasmid)
            sys.exit(f"ERROR: wrong 'plasmid' parameter! The plasmid name mustn't contain the word 'plasmid'.")
        elif(PLASMID_NAME_PATTERN.fullmatch(plasmid) is None and PLASMID_UNNAMED_PATTERN.fullmatch(plasmid) is None):
            log.error("Wrong 'plasmid' name! plasmid=%s", plasmid)
            sys.exit(f"ERROR: wrong 'plasmid' name! Plasmid names must either be named as 'unnamed', 'unnamed1', ... or start with a lower 'p', contain only digits, dots, underscores and letters, and are limited to 20 characters in total.")
    log.info('plasmid=%s', plasmid)
    taxon = ' '.join([t for t in [genus, species, strain] if t is not None])
    if(taxon == ''):
        taxon = None

    # annotation configurations
    global complete, prodigal_tf, translation_table, keep_sequence_headers, locus, locus_tag, locus_tag_increment, gram, replicons, compliant, user_proteins, user_hmms, meta, partial, regions
    complete = args.complete
    log.info('complete=%s', complete)
    prodigal_tf = args.prodigal_tf
    if(prodigal_tf is not None):
        try:
            if(prodigal_tf == ''):
                raise ValueError('File path argument must be non-empty')
            prodigal_tf_path = Path(args.prodigal_tf).resolve()
            check_readability('prodigal training', prodigal_tf_path)
            check_content_size('prodigal training', prodigal_tf_path)
            prodigal_tf = prodigal_tf_path
        except:
            log.error('provided prodigal training file not valid! path=%s', prodigal_tf)
            sys.exit(f'ERROR: Prodigal training file ({prodigal_tf}) not valid!')
    log.info('prodigal_tf=%s', prodigal_tf)
    translation_table = args.translation_table
    log.info('translation_table=%s', translation_table)
    gram = args.gram
    log.info('gram=%s', gram)
    compliant = args.compliant
    log.info('compliant=%s', compliant)
    if(compliant):
        min_sequence_length = 200
        log.info('compliant mode! min_contig_length=%s', min_sequence_length)
    meta = args.meta
    log.info('meta=%s', meta)
    partial = args.partial
    log.info('partial=%s', partial)
    locus = args.locus
    if(locus is not None):
        if(locus == ''):
            log.error("Empty 'locus' parameter! locus=%s", locus)
            sys.exit(f"ERROR: empty 'locus' parameter!")
        if(' ' in locus):
            log.error("Whitespace character in 'locus' parameter! locus=%s", locus)
            sys.exit(f"ERROR: whitespace character ({locus}) in 'locus' parameter!")
        if(bc.RE_INSDC_ID_PREFIX.fullmatch(locus) is None):
            log.error("Invalid 'locus' parameter! locus=%s", locus)
            sys.exit(f"ERROR: invalid 'locus' parameter ({locus})!\nLocus prefixes must contain between 1 and 20 alphanumeric or '-_' characters.")
    log.info('locus=%s', locus)
    locus_tag = args.locus_tag
    if(locus_tag is not None):
        if(locus_tag == ''):
            log.error("Empty 'locus-tag' parameter! locus=%s", locus_tag)
            sys.exit(f"ERROR: empty 'locus-tag' parameter!")
        if(' ' in locus_tag):
            log.error("Whitespace character in 'locus-tag' parameter! locus-tag=%s", locus_tag)
            sys.exit(f"ERROR: whitespace character ({locus_tag}) in 'locus-tag' parameter!")
        if(compliant):
            if(bc.RE_INSDC_LOCUSTAG_PREFIX.fullmatch(locus_tag) is None):
                log.error("INSDC-incompliant 'locus-tag' parameter! locus-tag=%s", locus_tag)
                sys.exit(f"ERROR: INSDC-incompliant 'locus-tag' parameter ({locus_tag})!\nINSDC Locus tag prefixes must contain between 3 and 12 alphanumeric uppercase characters and start with a letter.")
        else:
            if(bc.RE_LOCUSTAG_PREFIX.fullmatch(locus_tag) is None):
                log.error("Invalid 'locus-tag' parameter! locus-tag=%s", locus_tag)
                sys.exit(f"ERROR: invalid 'locus-tag' parameter ({locus_tag})!\nLocus tag prefixes must contain between 1 and 24 alphanumeric characters or '_.-' signs.")
    log.info('locus-tag=%s', locus_tag)
    locus_tag_increment = args.locus_tag_increment
    log.info('locus-tag-increment=%s', locus_tag_increment)
    keep_sequence_headers = args.keep_contig_headers
    log.info('keep_contig_headers=%s', keep_sequence_headers)
    replicons = args.replicons
    if(replicons is not None):
        try:
            if(replicons == ''):
                raise ValueError('File path argument must be non-empty')
            replicon_table_path = Path(args.replicons).resolve()
            check_readability('replicon table', replicon_table_path)
            check_content_size('replicon table', replicon_table_path)
            replicons = replicon_table_path
        except:
            log.error('provided replicon file not valid! path=%s', replicons)
            sys.exit(f'ERROR: replicon table file ({replicons}) not valid!')
    log.info('replicon-table=%s', replicons)
    user_proteins = check_user_proteins(args)
    user_hmms = args.hmms
    if(user_hmms is not None):
        try:
            if(user_hmms == ''):
                raise ValueError('File path argument must be non-empty')
            user_hmms_path = Path(user_hmms).resolve()
            check_readability('HMM', user_hmms_path)
            check_content_size('HMM', user_hmms_path)
            user_hmms = user_hmms_path
        except:
            log.error('provided HMM file not valid! path=%s', user_hmms)
            sys.exit(f'ERROR: HMM file ({user_hmms}) not valid!')

    regions = args.regions
    if(regions is not None):
        try:
            if(regions == ''):
                raise ValueError('File path argument must be non-empty')
            regions_path = Path(args.regions).resolve()
            check_readability('regions', regions_path)
            check_content_size('regions', regions_path)
            regions = regions_path
        except:
            log.error('provided regions file not valid! path=%s', regions)
            sys.exit(f'ERROR: regions file ({regions}) not valid!')
    log.info('regions=%s', regions)
    

    # workflow configurations
    global skip_trna, skip_tmrna, skip_rrna, skip_ncrna, skip_ncrna_region, skip_crispr, skip_cds, skip_pseudo, skip_sorf, skip_gap, skip_ori, skip_filter, skip_plot
    skip_trna = args.skip_trna
    log.info('skip-tRNA=%s', skip_trna)
    skip_tmrna = args.skip_tmrna
    log.info('skip-tmRNA=%s', skip_tmrna)
    skip_rrna = args.skip_rrna
    log.info('skip-rRNA=%s', skip_rrna)
    skip_ncrna = args.skip_ncrna
    log.info('skip-ncRNA=%s', skip_ncrna)
    skip_ncrna_region = args.skip_ncrna_region
    log.info('skip-ncRNA-region=%s', skip_ncrna_region)
    skip_crispr = args.skip_crispr
    log.info('skip-CRISPR=%s', skip_crispr)
    skip_cds = args.skip_cds
    log.info('skip-CDS=%s', skip_cds)
    skip_pseudo = args.skip_pseudo
    log.info('skip-pseudo=%s', skip_pseudo)
    skip_sorf = args.skip_sorf
    log.info('skip-sORF=%s', skip_sorf)
    skip_gap = args.skip_gap
    log.info('skip-gap=%s', skip_gap)
    skip_ori = args.skip_ori
    log.info('skip-ori=%s', skip_ori)
    skip_filter = args.skip_filter
    log.info('skip-filter=%s', skip_filter)
    skip_plot = args.skip_plot
    log.info('skip-plot=%s', skip_plot)


def check_readability(file_name: str, file_Path: Path):
    if(not os.access(str(file_Path), os.R_OK)):
        log.error('%s file not readable! path=%s', file_name, file_Path)
        sys.exit(f'ERROR: {file_name} file ({file_Path}) not readable!')


def check_content_size(file_name: str, file_path: Path):
    if(file_path.stat().st_size == 0):
        log.error('empty %s file! path=%s', file_name, file_path)
        sys.exit(f'ERROR: {file_name} file ({file_path}) is empty!')


def check_threads(args: Namespace) -> int:
    global threads
    threads = args.threads

    try:
        max_threads = len(os.sched_getaffinity(0))
        log.debug(f"max-threads={max_threads}")
    except AttributeError:
        max_threads = mp.cpu_count()
        log.debug(f"scheduler affinity not availabe! max-threads={max_threads}")

    if(threads == 0):
        threads = max_threads
        log.debug("request max threads.")
    elif(threads < 0):
        log.error("wrong argument for 'threads' parameter! threads=%i", threads)
        sys.exit(f"ERROR: wrong argument ({threads}) for 'threads' parameter! Value must be larger than/equal to 0.")
    elif(threads > max_threads):
        log.error("wrong argument for 'threads' parameter! More threads requested than available: requested=%i, available=%i", threads, max_threads)
        sys.exit(f"ERROR: wrong argument ({threads}) for 'threads' parameter! More threads requested ({threads}) than available ({max_threads}).")
    log.info('threads=%i', threads)
    return threads


def check_output_path(output: str, force_override: bool) -> Path:
    """Check provided output path
    Args:
        output (string): The output directory destination path
        force_override (Bool): Whether to override existing output directories
    """
    global output_path
    output_path = Path(output)
    if(not output_path.exists()):
        try:
            output_path.mkdir(parents=True, exist_ok=True)
        except:
            sys.exit(f'ERROR: could not resolve or create output directory ({output})!')
    else:
        if(output_path == Path(os.getcwd())):
            pass
        elif(force_override is False):
            sys.exit(f'ERROR: output path ({output_path}) already exists! Either provide a non-existent new path or force overwriting it via \'--force\'')
        elif(not os.access(str(output_path), os.X_OK)):
            sys.exit(f'ERROR: output path ({output_path}) not accessible!')
        elif(not os.access(str(output_path), os.W_OK)):
            sys.exit(f'ERROR: output path ({output_path}) not writable!')
    output_path = output_path.resolve()
    return output_path


def check_db_path(args: Namespace) -> Path:
    global db_path
    env = os.environ.copy()
    if(args.db):
        db_dir = args.db
        log.debug('test parameter db: db_tmp=%s', db_dir)
        try:
            db_tmp_path = Path(db_dir).resolve()
            if(db_tmp_path.is_dir()):
                db_path = db_tmp_path
                log.info('database: type=parameter, path=%s', db_path)
            else:
                log.error('unvalid database path: type=parameter, path=%s', db_tmp_path)
                raise IOError()
        except:
            sys.exit(f'ERROR: wrong database path! --db={db_dir}')
    elif('BAKTA_DB' in env):
        db_dir = env['BAKTA_DB']
        log.debug('test env db: db_tmp=%s', db_dir)
        try:
            db_tmp_path = Path(db_dir).resolve()
            if(db_tmp_path.is_dir()):
                db_path = db_tmp_path
                log.info('database: type=environment, path=%s', db_path)
            else:
                log.error('unvalid database path: type=environment, path=%s', db_tmp_path)
                raise IOError()
        except:
            sys.exit(f'ERROR: wrong database path! BAKTA_DB={db_dir}')
    else:
        base_dir = Path(__file__).parent
        db_tmp_path = base_dir.joinpath('db')
        log.debug('test base_dir db: db_tmp=%s', db_tmp_path)
        if(db_tmp_path.is_dir()):
            db_path = db_tmp_path
            log.info('database: type=base-dir, path=%s', db_path)
        else:
            log.error('unvalid database path: type=base-dir, path=%s', db_tmp_path)
            sys.exit('ERROR: database neither provided nor auto-detected!\nPlease, download the mandatory db and provide it via either the --db parameter, a BAKTA_DB environment variable or copy it into the Bakta base directory.\nFor further information please read the readme.md')
    return db_path


def check_user_proteins(args: Namespace):
    global user_proteins
    user_proteins = args.proteins
    if(user_proteins is not None):
        try:
            if(user_proteins == ''):
                raise ValueError('File path argument must be non-empty')
            user_proteins_path = Path(args.proteins).resolve()
            check_readability('user proteins', user_proteins_path)
            check_content_size('user proteins', user_proteins_path)
            user_proteins = user_proteins_path
            log.info('user-proteins=%s', user_proteins)
            return user_proteins
        except:
            log.error('provided user proteins file not valid! path=%s', user_proteins)
            sys.exit(f'ERROR: user proteins file ({user_proteins}) not valid!')
    else:
        return None


def check_tmp_path(args: Namespace) -> Path:
    global tmp_path
    if(args.tmp_dir is not None):
        tmp_path = Path(args.tmp_dir)
        if(not tmp_path.exists()):
            log.debug('dedicated temp dir does not exist! tmp-dir=%s', tmp_path)
            sys.exit(f'ERROR: dedicated temporary directory ({tmp_path}) does not exist!')
        else:
            log.info('use dedicated temp dir: path=%s', tmp_path)
            tmp_path = Path(tempfile.mkdtemp(dir=str(tmp_path))).resolve()
    else:
        tmp_path = Path(tempfile.mkdtemp()).resolve()
    log.info('tmp-path=%s', tmp_path)
    return tmp_path


import logging

from collections import OrderedDict
from pathlib import Path
from typing import Dict, Sequence

import pyhmmer
from pyhmmer.easel import AA, DigitalSequenceBlock, TextSequenceBlock, TextSequence

import bakta.config as cfg
import bakta.constants as bc


log = logging.getLogger('ORF')


def detect_spurious(orfs: Sequence[dict]):
    """Detect spurious ORFs with AntiFam"""
    discarded_orfs = []
    orf_by_aa_digest = get_orf_dictionary(orfs)
    alphabet: "AA" = pyhmmer.easel.Alphabet.amino()
    proteins: "DigitalSequenceBlock[AA]" = TextSequenceBlock(TextSequence(sequence=orf['aa'], name=get_orf_key(orf)) for orf in orfs).digitize(alphabet)
    with pyhmmer.plan7.HMMFile(cfg.db_path.joinpath('antifam'), alphabet=alphabet) as hmm_fh:
        for top_hits in pyhmmer.hmmsearch(hmm_fh, proteins, bit_cutoffs='gathering', cpus=cfg.threads):
            for hit in top_hits:
                orf = orf_by_aa_digest[hit.name]
                if hit.evalue > bc.MIN_HMM_EVALUE:
                    log.debug(
                        'discard low spurious E value: seq=%s, start=%i, stop=%i, strand=%s, subject=%s, evalue=%1.1e, bitscore=%f',
                        orf['sequence'], orf['start'], orf['stop'], orf['strand'], hit.best_domain.alignment.hmm_name, hit.evalue, hit.score
                    )
                else:
                    discard = OrderedDict()
                    discard['type'] = bc.DISCARD_TYPE_SPURIOUS
                    discard['description'] = f'(partial) homology to spurious sequence HMM (AntiFam:{hit.best_domain.alignment.hmm_accession})'
                    discard['score'] = hit.score
                    discard['evalue'] = hit.evalue

                    orf['discarded'] = discard
                    discarded_orfs.append(orf)
                    log.info(
                        'discard spurious: seq=%s, start=%i, stop=%i, strand=%s, homology=%s, evalue=%1.1e, bitscore=%f',
                        orf['sequence'], orf['start'], orf['stop'], orf['strand'], hit.best_domain.alignment.hmm_name, hit.evalue, hit.score
                    )
    log.info('discarded=%i', len(discarded_orfs))
    return discarded_orfs


def get_orf_key(orf: dict) -> str:
    """Generate a standardized and unique ORF-like feature key for internal store/analyze/parse/retrieval cycles."""
    return f"{orf['aa_hexdigest']}-{orf['sequence']}-{orf['start']}-{orf['stop']}-{orf['strand']}-{orf.get('source', 'internal')}"


def get_orf_dictionary(orfs: Sequence[dict]) -> Dict[str, dict]:
    """create a standardized ORF-like feature dict for internal store/analyze/parse/retrieval cycles."""
    return {get_orf_key(orf): orf for orf in orfs}


def write_internal_faa(features: Sequence[dict], faa_path: Path):
    """Write aa sequences to internal temporary Fasta file."""
    log.info('write internal aa seqs: # seqs=%i, path=%s', len(features), faa_path)
    with faa_path.open(mode='wt') as fh:
        for orf in features:
            fh.write(f">{get_orf_key(orf)}\n{orf['aa']}\n")
    



import logging

from pathlib import Path
from types import LambdaType
from typing import Dict, Sequence

import bakta
import bakta.config as cfg
import bakta.constants as bc
import bakta.ips as bips
import bakta.ups as bups
import bakta.psc as bpsc
import bakta.pscc as bpscc

from bakta.features.annotation import calc_annotation_score

log = logging.getLogger('TSV')


def write_features(sequences: Sequence[dict], features_by_sequence: Dict[str, dict], tsv_path: Path):
    """Export features in TSV format."""
    log.info('write feature tsv: path=%s', tsv_path)

    with tsv_path.open('wt') as fh:
        fh.write('# Annotated with Bakta\n')
        fh.write(f'# Software: v{cfg.version}\n')
        fh.write(f"# Database: v{cfg.db_info['major']}.{cfg.db_info['minor']}, {cfg.db_info['type']}\n")
        fh.write(f'# DOI: {bc.BAKTA_DOI}\n')
        fh.write(f'# URL: {bc.BAKTA_URL}\n')
        fh.write('#Sequence Id\tType\tStart\tStop\tStrand\tLocus Tag\tGene\tProduct\tDbXrefs\n')

        for seq in sequences:
            for feat in features_by_sequence[seq['id']]:
                seq_id = feat['sequence'] if 'sequence' in feat else feat['contig']  # <1.10.0 compatibility
                feat_type = feat['type']
                if(feat_type == bc.FEATURE_GAP):
                    feat_type = bc.INSDC_FEATURE_ASSEMBLY_GAP if feat['length'] >= 100 else bc.INSDC_FEATURE_GAP

                gene = feat['gene'] if feat.get('gene', None) else ''
                product = feat.get('product', '')
                if(bc.PSEUDOGENE in feat):
                    product = f"(pseudo) {product}"
                elif(feat.get('truncated', '') == bc.FEATURE_END_5_PRIME):
                    product = f"(5' truncated) {product}"
                elif(feat.get('truncated', '') == bc.FEATURE_END_3_PRIME):
                    product = f"(3' truncated) {product}"
                elif(feat.get('truncated', '') == bc.FEATURE_END_BOTH):
                    product = f"(partial) {product}"
                fh.write('\t'.join(
                    [
                        seq_id,
                        feat_type,
                        str(feat['start']),
                        str(feat['stop']),
                        feat['strand'],
                        feat.get('locus', ''),
                        gene,
                        product,
                        ', '.join(sorted(feat.get('db_xrefs', [])))
                    ])
                )
                fh.write('\n')
                if(feat_type == bc.FEATURE_CRISPR):
                    i = 0
                    while i < len(feat['spacers']):
                        repeat = feat['repeats'][i]
                        fh.write('\t'.join([seq_id, bc.FEATURE_CRISPR_REPEAT, str(repeat['start']), str(repeat['stop']), repeat['strand'], '', '', f"CRISPR repeat", '']))
                        fh.write('\n')
                        spacer = feat['spacers'][i]
                        fh.write('\t'.join([seq_id, bc.FEATURE_CRISPR_SPACER, str(spacer['start']), str(spacer['stop']), spacer['strand'], '', '', f"CRISPR spacer, sequence {spacer['sequence']}", '']))
                        fh.write('\n')
                        i += 1
                    if(len(feat['repeats']) - 1 == i):
                        repeat = feat['repeats'][i]
                        fh.write('\t'.join([seq_id, bc.FEATURE_CRISPR_REPEAT, str(repeat['start']), str(repeat['stop']), repeat['strand'], '', '', f"CRISPR repeat", '']))
                        fh.write('\n')
    return


def write_feature_inferences(sequences: Sequence[dict], features_by_sequence: Dict[str, dict], tsv_path: Path):
    """Export feature inference statistics in TSV format."""
    log.info('write tsv: path=%s', tsv_path)

    with tsv_path.open('wt') as fh:
        fh.write('# Annotated with Bakta\n')
        fh.write(f'# Software: v{cfg.version}\n')
        fh.write(f"# Database: v{cfg.db_info['major']}.{cfg.db_info['minor']}, {cfg.db_info['type']}\n")
        fh.write(f'# DOI: {bc.BAKTA_DOI}\n')
        fh.write(f'# URL: {bc.BAKTA_URL}\n')
        fh.write('#Sequence Id\tType\tStart\tStop\tStrand\tLocus Tag\tScore\tEvalue\tQuery Cov\tSubject Cov\tId\tAccession\n')

        for seq in sequences:
            for feat in features_by_sequence[seq['id']]:
                if(feat['type'] in [bc.FEATURE_CDS, bc.FEATURE_SORF]):
                    score, evalue, query_cov, subject_cov, identity, accession = None, None, None, None, None, '-'
                    if('expert' in feat):
                        expert_hits = feat.get('expert', [])
                        top_expert_hit = sorted(expert_hits,key=lambda k: (k['rank'], k.get('score', 0), calc_annotation_score(k)), reverse=True)[0]
                        query_cov = top_expert_hit.get('query_cov', None)
                        subject_cov = top_expert_hit.get('subject_cov', None)
                        identity = top_expert_hit.get('identity', None)
                        score = top_expert_hit.get('score', None)
                        evalue = top_expert_hit.get('evalue', None)
                        accession = [dbxref for dbxref in top_expert_hit.get('db_xrefs', []) if dbxref.split(':')[0] in [bc.DB_XREF_UNIREF, bc.DB_XREF_UNIPARC, bc.DB_XREF_IS, bc.DB_XREF_NCBI_PROTEIN, bc.DB_XREF_NCBI_FAMILIES, bc.DB_XREF_NCBI_BLASTRULES, bc.DB_XREF_VFDB, bc.DB_XREF_USER_PROTEINS, bc.DB_XREF_USER_HMMS]][0]
                    elif('ups' in feat or 'ips' in feat):
                        query_cov = 1
                        subject_cov = 1
                        identity = 1
                        evalue = 0
                        accession = f"{bc.DB_XREF_UNIREF}:{feat['ips'][bips.DB_IPS_COL_UNIREF100]}" if 'ips' in feat else f"{bc.DB_XREF_UNIPARC}:{feat['ups'][bups.DB_UPS_COL_UNIPARC]}"
                    elif('psc' in feat or 'pscc' in feat):
                        psc_type = 'psc' if 'psc' in feat else 'pscc'
                        query_cov = feat[psc_type]['query_cov']
                        subject_cov = feat[psc_type].get('subject_cov', None)
                        identity = feat[psc_type]['identity']
                        score = feat[psc_type].get('score', None)
                        evalue = feat[psc_type].get('evalue', None)
                        accession = f"{bc.DB_XREF_UNIREF}:{feat['psc'][bpsc.DB_PSC_COL_UNIREF90]}" if 'psc' in feat else f"{bc.DB_XREF_UNIREF}:{feat['pscc'][bpscc.DB_PSCC_COL_UNIREF50]}"
                    fh.write('\t'.join(
                        [
                            feat['sequence'] if 'sequence' in feat else feat['contig'],  # <1.10.0 compatibility
                            feat['type'],
                            str(feat['start']),
                            str(feat['stop']),
                            feat['strand'],
                            feat['locus'],
                            f"{score:0.1f}" if score != None else '-',
                            ('0.0' if evalue == 0 else f"{evalue:1.1e}") if evalue != None else '-',
                            ('1.0' if query_cov == 1 else f"{query_cov:0.3f}") if query_cov != None else '-',
                            ('1.0' if subject_cov == 1 else f"{subject_cov:0.3f}") if subject_cov != None else '-',
                            ('1.0' if identity == 1 else f"{identity:0.3f}") if identity != None else '-',
                            accession
                        ])
                    )
                    fh.write('\n')
                elif(feat['type'] in [bc.FEATURE_T_RNA, bc.FEATURE_R_RNA, bc.FEATURE_NC_RNA, bc.FEATURE_NC_RNA_REGION]):
                    accession = '-' if feat['type'] == bc.FEATURE_T_RNA else [xref for xref in feat['db_xrefs'] if bc.DB_XREF_RFAM in xref][0]
                    fh.write('\t'.join(
                        [
                            feat['sequence'] if 'sequence' in feat else feat['contig'],  # <1.10.0 compatibility
                            feat['type'],
                            str(feat['start']),
                            str(feat['stop']),
                            feat['strand'],
                            feat['locus'] if 'locus' in feat else '-',
                            f"{feat['score']:0.1f}",
                            ('0.0' if feat['evalue'] == 0 else f"{feat['evalue']:1.1e}") if 'evalue' in feat else '-',
                            ('1.0' if feat['query_cov'] == 1 else f"{feat['query_cov']:0.3f}") if 'query_cov' in feat else '-',
                            ('1.0' if feat['subject_cov'] == 1 else f"{feat['subject_cov']:0.3f}") if 'subject_cov' in feat else '-',
                            ('1.0' if feat['identity'] == 1 else f"{feat['identity']:0.3f}") if 'identity' in feat else '-',
                            accession
                        ])
                    )
                    fh.write('\n')
    return


def write_protein_features(features: Sequence[dict], header_columns: Sequence[str], mapping: LambdaType, tsv_path: Path):
    """Export protein features in TSV format."""
    log.info('write protein feature tsv: path=%s', tsv_path)

    with tsv_path.open('wt') as fh:
        fh.write(f'#Annotated with Bakta (v{cfg.version}): https://github.com/oschwengers/bakta\n')
        fh.write(f"#Database (v{cfg.db_info['major']}.{cfg.db_info['minor']}): https://doi.org/10.5281/zenodo.4247252\n")
        fh.write('\t'.join(header_columns))
        fh.write('\n')
        for feat in features:
            columns = mapping(feat)
            fh.write('\t'.join(columns))
            fh.write('\n')
    return


def write_hypotheticals(hypotheticals: Sequence[dict], tsv_path: Path):
    """Export hypothetical information in TSV format."""
    log.info('write hypothetical tsv: path=%s', tsv_path)

    with tsv_path.open('wt') as fh:
        fh.write(f'#Annotated with Bakta v{cfg.version}, https://github.com/oschwengers/bakta\n')
        fh.write(f"#Database v{cfg.db_info['major']}.{cfg.db_info['minor']}, https://doi.org/10.5281/zenodo.4247252\n")
        fh.write('#Sequence Id\tStart\tStop\tStrand\tLocus Tag\tMol Weight [kDa]\tIso El. Point\tPfam hits\tDbxrefs\n')
        for hypo in hypotheticals:
            pfams = [f"{pfam['id']}|{pfam['name']}" for pfam in hypo.get('pfams', [])]
            seq_stats = hypo['seq_stats']
            mol_weight = f"{(seq_stats['molecular_weight']/1000):.1f}" if seq_stats['molecular_weight'] else 'NA'
            iso_point = f"{seq_stats['isoelectric_point']:.1f}" if seq_stats['isoelectric_point'] else 'NA'
            seq_id = hypo['sequence'] if 'sequence' in hypo else hypo['contig']  # <1.10.0 compatibility
            fh.write(f"{seq_id}\t{hypo['start']}\t{hypo['stop']}\t{hypo['strand']}\t{hypo.get('locus', '')}\t{mol_weight}\t{iso_point}\t{', '.join(sorted(pfams))}\t{', '.join(sorted(hypo.get('db_xrefs', [])))}\n")
    return



import logging
import re

from collections import defaultdict
from datetime import date
from pathlib import Path
from typing import Sequence, Tuple

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation, CompoundLocation, AfterPosition, BeforePosition

import bakta.config as cfg
import bakta.constants as bc
import bakta.features.annotation as ba
import bakta.so as so


log = logging.getLogger('INSDC')


def build_biopython_sequence_list(data: dict, features: Sequence[dict]):
    sequence_feature_map = defaultdict(list)
    if len(features) > 0:
        sequence_id_key = 'sequence' if 'sequence' in features[0] else 'contig'  # <1.10.0 compatibility
        for feature in features:
            sequence_feature_map[feature[sequence_id_key]].append(feature)
    sequence_list = []
    for seq in data['sequences']:
        sequence_features = sequence_feature_map[seq['id']]
        comment = (
            'Annotated with Bakta',
            f"Software: v{cfg.version}\n",
            f"Database: v{cfg.db_info['major']}.{cfg.db_info['minor']}, {cfg.db_info['type']}\n",
            f'DOI: {bc.BAKTA_DOI}\n',
            f'URL: {bc.BAKTA_URL}\n',
            '\n',
            '##Genome Annotation Summary:##\n',
            f"{'Annotation Date':<30} :: {cfg.run_end.strftime('%m/%d/%Y, %H:%M:%S')}\n",
            f"{'CDSs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_CDS or feat['type'] == bc.FEATURE_SORF]):5,}\n",
            f"{'tRNAs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_T_RNA]):5,}\n",
            f"{'tmRNAs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_TM_RNA]):5,}\n",
            f"{'rRNAs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_R_RNA]):5,}\n",
            f"{'ncRNAs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_NC_RNA]):5,}\n",
            f"{'regulatory ncRNAs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_NC_RNA_REGION]):5,}\n",
            f"{'CRISPR Arrays':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_CRISPR]):5,}",
            f"{'oriCs/oriVs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_ORIC or feat['type'] == bc.FEATURE_ORIV]):5,}",
            f"{'oriTs':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_ORIT]):5,}",
            f"{'gaps':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_GAP]):5,}",
            f"{'pseudogenes':<30} :: {len([feat for feat in sequence_features if feat['type'] == bc.FEATURE_CDS and bc.PSEUDOGENE in feat]):5,}\n"
        )
        sequence_annotations = {
            'molecule_type': 'DNA',
            'source': data['genome'].get('taxon', ''),
            'date': cfg.run_end.strftime('%d-%b-%Y').upper(),
            'topology': seq['topology'],
            'data_file_division': 'HGT' if seq['type'] == bc.REPLICON_CONTIG else 'BCT',
            # 'accession': '*',  # hold back until EMBL output bug is fixed in BioPython (https://github.com/biopython/biopython/pull/3572)
            'comment': comment
            # TODO: taxonomy
        }
        source_qualifiers = {
            'mol_type': 'genomic DNA'
            # 'molecule_type': 'DNA' #  might be necessary in BioPython > 1.78 along with removal of Seq(..., generic_dna)
        }

        description = ''
        if(data['genome'].get('taxon', None)):
            sequence_annotations['organism'] = data['genome']['taxon']
            source_qualifiers['organism'] = data['genome']['taxon']
            description = data['genome']['taxon']
        if(data['genome']['strain']):
            source_qualifiers['strain'] = data['genome']['strain']

        if(seq['type'] == bc.REPLICON_PLASMID):
            source_qualifiers['plasmid'] = seq['name'] if seq.get('name', None) else 'unnamed'
            description = f"{description} plasmid {seq.get('name', 'unnamed')}"
            description += ', complete sequence' if seq['complete'] else ', whole genome shotgun sequence'
        elif(seq['type'] == bc.REPLICON_CHROMOSOME):
            if seq.get('name', None):
                source_qualifiers['chromosome'] = seq['name']
            description = f'{description} chromosome, complete genome' if seq['complete'] else f"{description} chromosome {seq['id']}, whole genome shotgun sequence"
        else:
            description += f" {seq['id']}, whole genome shotgun sequence"

        if(len(description) > 0 and description[0] == ' '):  # discard potential leading whitespace
            description = description[1:]

        seq_bio = Seq(seq['nt']) if 'nt' in seq else Seq(seq['sequence'])  # <1.10.0 compatibility
        sequence_record = SeqIO.SeqRecord(id=seq['id'], name=seq['id'], description=description, annotations=sequence_annotations, seq=seq_bio)

        source = SeqFeature(FeatureLocation(0, seq['length'], strand=+1), type='source', qualifiers=source_qualifiers)
        seq_feature_list = [source]

        for feature in sequence_features:
            insdc_feature_type = None
            qualifiers = {
                'note': []
            }
            if('db_xrefs' in feature):
                if(cfg.compliant):
                    qualifiers['db_xref'], qualifiers['note'] = revise_dbxref_insdc(feature['db_xrefs'])
                else:
                    qualifiers['db_xref'] = feature['db_xrefs']
            if('product' in feature):
                qualifiers['product'] = feature['product']
            if('locus' in feature):
                qualifiers['locus_tag'] = feature['locus']

            accompanying_features=[]
            if(feature['type'] == bc.FEATURE_GAP):
                insdc_feature_type = bc.INSDC_FEATURE_GAP
                qualifiers['estimated_length'] = feature['length']
            elif(feature['type'] == bc.FEATURE_ORIC or feature['type'] == bc.FEATURE_ORIV):
                # TODO: Add fuzzy positions for oriC/oriV
                insdc_feature_type = bc.INSDC_FEATURE_ORIGIN_REPLICATION
                qualifiers['inference'] = 'similar to DNA sequence'
                qualifiers['note'].append(feature['product'])
                if('product' in qualifiers):
                    qualifiers['note'] = feature['product']
                    del qualifiers['product']
            elif(feature['type'] == bc.FEATURE_ORIT):
                # TODO: Add fuzzy positions for oriT
                insdc_feature_type = bc.INSDC_FEATURE_ORIGIN_TRANSFER
                qualifiers['inference'] = 'similar to DNA sequence'
                qualifiers['note'].append(feature['product'])
                if('product' in qualifiers):
                    del qualifiers['product']
            elif(feature['type'] == bc.FEATURE_CDS) or (feature['type'] == bc.FEATURE_SORF):
                if(bc.PSEUDOGENE in feature):
                    qualifiers[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNPROCESSED if feature[bc.PSEUDOGENE]['paralog'] else bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNITARY
                    qualifiers['note'].append(feature[bc.PSEUDOGENE]['description'])
                else:
                    qualifiers['protein_id'] = f"gnl|Bakta|{feature['locus']}"
                    qualifiers['translation'] = feature['aa']
                qualifiers['codon_start'] = 1
                qualifiers['transl_table'] = cfg.translation_table
                insdc_feature_type = bc.INSDC_FEATURE_CDS
                inference = []
                if(feature['type'] == bc.FEATURE_CDS):
                    if(feature.get('source', None) == bc.CDS_SOURCE_USER):
                        inference.append('EXISTENCE:non-experimental evidence, no additional details recorded')
                    else:
                        inference.append('ab initio prediction:Prodigal:2.6')
                else:
                    inference.append(f"ab initio prediction:Bakta:{'.'.join(cfg.version.split('.')[0:2])}")
                if('ncbi_nrp_id' in feature.get('ups', {})):
                    nrp_id = feature['ups']['ncbi_nrp_id']
                    inference.append(f'similar to AA sequence:{bc.DB_XREF_REFSEQ_NRP}:{nrp_id}')
                elif('uniref100_id' in feature.get('ips', {})):
                    ips_subject_id = feature['ips']['uniref100_id']
                    inference.append(f'similar to AA sequence:{bc.DB_XREF_UNIREF}:{ips_subject_id}')
                elif('uniparc_id' in feature.get('ups', {})):
                    uniparc_id = feature['ups']['uniparc_id']
                    inference.append(f'similar to AA sequence:{bc.DB_XREF_UNIPARC}:{uniparc_id}')
                elif('uniref90_id' in feature.get('psc', {}) and feature.get('psc', {}).get('valid', False)):
                    psc_subject_id = feature['psc']['uniref90_id']
                    inference.append(f'similar to AA sequence:{bc.DB_XREF_UNIREF}:{psc_subject_id}')
                elif('uniref50_id' in feature.get('pscc', {})):
                    pscc_subject_id = feature['pscc']['uniref50_id']
                    inference.append(f'similar to AA sequence:{bc.DB_XREF_UNIREF}:{pscc_subject_id}')
                qualifiers['inference'] = inference
                if(cfg.compliant):
                    qualifiers['note'], ec_number = extract_ec_from_notes_insdc(qualifiers, 'note')
                    if(ec_number is not None):
                            qualifiers['EC_number'] = ec_number
                if('exception' in feature):
                    ex = feature['exception']
                    if(ex['type'] == 'selenocysteine'):
                        pos = f"{ex['start']}..{ex['stop']}"
                        if(feature['strand'] == bc.STRAND_REVERSE):
                            pos = f"complement({pos})"
                        qualifiers['transl_except']=f"(pos:{pos},aa:{ex['aa']})"
                        qualifiers['note'].append(f"codon on position {ex['codon_position']} is a {ex['type']} codon")
                    elif(ex['type'] == 'ribosomal_slippage'):
                        qualifiers[bc.INSDC_FEAUTRE_CDS_RIBOSOMAL_SLIPPAGE] = None
                if(bc.FEATURE_SIGNAL_PEPTIDE in feature):
                    sigpep_qualifiers = {}
                    sigpep_qualifiers['locus_tag'] = feature['locus']
                    sigpep_qualifiers['inference'] = 'ab initio prediction:DeepSig:1.2'
                    sigpep = feature[bc.FEATURE_SIGNAL_PEPTIDE]
                    sigpep_strand = 1 if feature['strand'] == bc.STRAND_FORWARD else -1 if feature['strand'] == bc.STRAND_REVERSE else 0
                    sigpep_location = FeatureLocation(sigpep['start'] - 1, sigpep['stop'], strand=sigpep_strand)
                    sigpep_feature = SeqFeature(sigpep_location, type=bc.INSDC_FEATURE_SIGNAL_PEPTIDE, qualifiers=sigpep_qualifiers)
                    accompanying_features.append(sigpep_feature)
            elif(feature['type'] == bc.FEATURE_T_RNA):
                if('amino_acid' in feature and 'anti_codon' in feature):
                    if('anti_codon_pos' in feature):
                        anti_codon_pos = feature['anti_codon_pos']
                        qualifiers['anticodon'] = f"(pos:{anti_codon_pos[0]}..{anti_codon_pos[1]},aa:{feature['amino_acid']},seq:{feature['anti_codon']})"
                    else:
                        qualifiers['note'].append(f"tRNA-{feature['amino_acid']} ({feature['anti_codon']})")
                qualifiers['inference'] = 'profile:tRNAscan:2.0'
                insdc_feature_type = bc.INSDC_FEATURE_T_RNA
                if(bc.PSEUDOGENE in feature):
                    qualifiers[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNKNOWN
            elif(feature['type'] == bc.FEATURE_TM_RNA):
                qualifiers['inference'] = 'profile:aragorn:1.2'
                if('tag' in feature):
                    qualifiers['tag_peptide'] = f"{feature['tag']['start']}..{feature['tag']['stop']}"
                insdc_feature_type = bc.INSDC_FEATURE_TM_RNA
            elif(feature['type'] == bc.FEATURE_R_RNA):
                for rfam_id in [dbxref.split(':')[1] for dbxref in feature['db_xrefs'] if dbxref.split(':')[0] == bc.DB_XREF_RFAM]:
                    qualifiers['inference'] = f'profile:Rfam:{rfam_id}'
                insdc_feature_type = bc.INSDC_FEATURE_R_RNA
            elif(feature['type'] == bc.FEATURE_NC_RNA):
                for rfam_id in [dbxref.split(':')[1] for dbxref in feature['db_xrefs'] if dbxref.split(':')[0] == bc.DB_XREF_RFAM]:
                    qualifiers['inference'] = f'profile:Rfam:{rfam_id}'
                qualifiers[bc.INSDC_FEATURE_NC_RNA_CLASS] = select_ncrna_class(feature)
                insdc_feature_type = bc.INSDC_FEATURE_NC_RNA
            elif(feature['type'] == bc.FEATURE_NC_RNA_REGION):
                for rfam_id in [dbxref.split(':')[1] for dbxref in feature['db_xrefs'] if dbxref.split(':')[0] == bc.DB_XREF_RFAM]:
                    qualifiers['inference'] = f'profile:Rfam:{rfam_id}'
                qualifiers[bc.INSDC_FEATURE_REGULATORY_CLASS] = select_regulatory_class(feature)
                insdc_feature_type = bc.INSDC_FEATURE_REGULATORY
                qualifiers['note'].append(feature['product'])
                qualifiers.pop('product', None)
            elif(feature['type'] == bc.FEATURE_CRISPR):
                qualifiers[bc.INSDC_FEATURE_REPEAT_FAMILY] = 'CRISPR'
                qualifiers[bc.INSDC_FEATURE_REPEAT_TYPE] = 'direct'
                qualifiers[bc.INSDC_FEATURE_REPEAT_UNIT_SEQ] = feature['repeat_consensus']
                qualifiers['inference'] = 'COORDINATES:alignment:pilercr:1.02'
                insdc_feature_type = bc.INSDC_FEATURE_REPEAT_REGION
                qualifiers['note'].append(feature['product'])
                qualifiers.pop('product', None)

            strand = None
            if(feature['strand'] == bc.STRAND_FORWARD):
                strand = 1
            elif(feature['strand'] == bc.STRAND_REVERSE):
                strand = -1
            elif(feature['strand'] == bc.STRAND_UNKNOWN):
                strand = 0

            start = feature['start'] - 1
            stop = feature['stop']
            if('edge' in feature):
                fl_1 = FeatureLocation(start, seq['length'], strand=strand)
                fl_2 = FeatureLocation(0, stop, strand=strand)
                if(feature['strand'] == bc.STRAND_REVERSE):
                    feature_location = CompoundLocation([fl_2, fl_1])
                else:
                    feature_location = CompoundLocation([fl_1, fl_2])
            else:
                if('truncated' in feature):
                    if(bc.PSEUDOGENE not in feature):  # only add /pseudo qualifier if /pseudogene is not already set
                        qualifiers[bc.INSDC_FEATURE_PSEUDO] = None
                    if(feature['truncated'] == bc.FEATURE_END_5_PRIME):
                        qualifiers['note'].append("(5' truncated)")
                        if(feature['strand'] == bc.STRAND_FORWARD):
                            start = BeforePosition(start)
                        else:
                            stop = AfterPosition(stop)
                    elif(feature['truncated'] == bc.FEATURE_END_3_PRIME):
                        qualifiers['note'].append("(3' truncated)")
                        if(feature['strand'] == bc.STRAND_FORWARD):
                            stop = AfterPosition(stop)
                        else:
                            start = BeforePosition(start)
                    elif(feature['truncated'] == bc.FEATURE_END_BOTH):
                        qualifiers['note'].append('(partial)')
                        start = BeforePosition(start)
                        stop = AfterPosition(stop)
                feature_location = FeatureLocation(start, stop, strand=strand)
            if(feature.get('locus', None)):
                gene_qualifier = {
                    'locus_tag': feature['locus']
                }
                if(feature.get('gene', None)):
                    if(cfg.compliant):
                        if(ba.RE_GENE_SYMBOL.fullmatch(feature['gene'])):  # discard non-standard gene symbols
                            qualifiers['gene'] = feature['gene']
                            gene_qualifier['gene'] = feature['gene']
                        else:
                            qualifiers.pop('gene', None)
                    else:
                        qualifiers['gene'] = feature['gene']
                        gene_qualifier['gene'] = feature['gene']
                if(bc.PSEUDOGENE in feature):
                    if(feature['type'] == bc.FEATURE_CDS):
                        gene_qualifier[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNPROCESSED if feature[bc.PSEUDOGENE]['paralog'] else bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNITARY
                    else:
                        gene_qualifier[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNKNOWN
                elif('truncated' in feature):
                    gene_qualifier[bc.INSDC_FEATURE_PSEUDO] = None
                gen_seqfeat = SeqFeature(feature_location, type='gene', qualifiers=gene_qualifier)
                seq_feature_list.append(gen_seqfeat)
            feat_seqfeat = SeqFeature(feature_location, type=insdc_feature_type, qualifiers=qualifiers)
            seq_feature_list.append(feat_seqfeat)
            for acc_feature in accompanying_features:  # add accompanying features, e.g. signal peptides
                seq_feature_list.append(acc_feature)
        sequence_record.features = seq_feature_list
        sequence_list.append(sequence_record)
    return sequence_list


def write_features(data: dict, features: Sequence[dict], genbank_output_path: Path, embl_output_path: Path):
    log.debug('prepare: genbank=%s, embl=%s', genbank_output_path, embl_output_path)

    sequence_list = build_biopython_sequence_list(data, features)
    with genbank_output_path.open('wt', encoding='utf-8') as fh:
        log.info('write GenBank: path=%s', genbank_output_path)
        SeqIO.write(sequence_list, fh, format='genbank')

    with embl_output_path.open('wt', encoding='utf-8') as fh:
        log.info('write EMBL: path=%s', embl_output_path)
        SeqIO.write(sequence_list, fh, format='embl')


def select_ncrna_class(feature: dict) -> str:
    feature_class = feature['class']
    if(feature_class is None):
        return bc.INSDC_FEATURE_NC_RNA_CLASS_OTHER
    else:
        if(isinstance(feature_class, list)):  #  workaround for JSON-imported features
            feature_class = so.SO(feature_class[0], feature_class[1])
            feature['class'] = feature_class

        if(feature_class.id == so.SO_NCRNA_GENE_ANTISENSE.id):
            return bc.INSDC_FEATURE_NC_RNA_CLASS_ANTISENSE
        elif(feature_class.id == so.SO_NCRNA_GENE_RIBOZYME.id):
            return bc.INSDC_FEATURE_NC_RNA_CLASS_RIBOZYME
        elif(feature_class.id == so.SO_NCRNA_GENE_RNASEP.id):
            return bc.INSDC_FEATURE_NC_RNA_CLASS_RNASEP
        else:
            return bc.INSDC_FEATURE_NC_RNA_CLASS_OTHER


def select_regulatory_class(feature: dict) -> str:
    feature_class = feature['class']
    if(feature_class is None):
        return bc.INSDC_FEATURE_REGULATORY_CLASS_OTHER
    else:
        if(isinstance(feature_class, list)):  #  workaround for JSON-imported features
            feature_class = so.SO(feature_class[0], feature_class[1])
            feature['class'] = feature_class

        if(feature_class.id == so.SO_CIS_REG_ATTENUATOR.id):
            return bc.INSDC_FEATURE_REGULATORY_CLASS_ATTENUATOR
        elif(feature_class.id == so.SO_CIS_REG_RIBOSWITCH.id):
            return bc.INSDC_FEATURE_REGULATORY_CLASS_RIBOSWITCH
        elif(feature_class.id == so.SO_CIS_REG_THERMOMETER.id):
            return bc.INSDC_FEATURE_REGULATORY_CLASS_RESPONSE_ELEMENT
        elif(feature_class.id == so.SO_CIS_REG_RECODING_STIMULATION_REGION.id or feature_class.id == so.SO_CIS_REG_FRAMESHIFT.id):
            return bc.INSDC_FEATURE_REGULATORY_CLASS_RECODING_STIMULATORY_REGION
        elif(feature_class.id == so.SO_CIS_REG_RIBOSOME_BINDING_SITE.id):
            return bc.INSDC_FEATURE_REGULATORY_CLASS_RIBOSOME_BINDING_SITE
        else:
            return bc.INSDC_FEATURE_REGULATORY_CLASS_OTHER


def revise_product_insdc(product: str):
    """Revise product name for INSDC compliant submissions"""

    old_product = product
    if(re.search(r'(uncharacteri[sz]ed)', product, flags=re.IGNORECASE)):  # replace putative synonyms)
        product = re.sub(r'(uncharacteri[sz]ed)', 'putative', product, flags=re.IGNORECASE)
        log.info('fix product: replace putative synonyms. new=%s, old=%s', product, old_product)

    old_product = product
    if(product.count('(') != product.count(')')):  # remove unbalanced parentheses
        product = product.replace('(', '').replace(')', '')  # ToDo: find and replace only legend parentheses
        log.info('fix product: remove unbalanced parantheses. new=%s, old=%s', product, old_product)

    old_product = product
    if(product.count('[') != product.count(']')):  # remove unbalanced brackets
        product = product.replace('[', '').replace(']', '')  # ToDo: find and replace only legend bracket
        log.info('fix product: remove unbalanced brackets. new=%s, old=%s', product, old_product)

    return product


def revise_dbxref_insdc(dbxrefs: Sequence[str]) -> Tuple[Sequence[str], Sequence[str]]:
    """Remove INSDC non-compliant DbXrefs."""
    insdc_valid_dbxrefs = [bc.DB_XREF_UNIPROTKB, bc.DB_XREF_GO, bc.DB_XREF_PFAM, bc.DB_XREF_RFAM]
    valid_dbxrefs = []
    invalid_dbxrefs = []
    for dbxref in dbxrefs:
        if(dbxref.split(':')[0] in insdc_valid_dbxrefs):
            valid_dbxrefs.append(dbxref)
        else:
            invalid_dbxrefs.append(dbxref)
    return valid_dbxrefs, invalid_dbxrefs


def extract_ec_from_notes_insdc(qualifiers: dict, note_key: str):
    note_tmp = []
    ec_number = None
    for note in qualifiers[note_key]:  # move EC numbers from note to EC_number qualifier
        if(note.split(':')[0] == bc.DB_XREF_EC):
            ec_number = note.replace('EC:', '')
        else:
            note_tmp.append(note)
    return note_tmp, ec_number



import logging
import re

from collections import OrderedDict
from typing import Sequence

import bakta.constants as bc


log = logging.getLogger('GAP')
RE_ASSEMBLY_GAP = re.compile(r'N{1,}', flags=0)


def detect_assembly_gaps(data: dict) -> Sequence[dict]:
    gaps = []
    for seq in data['sequences']:
        m = RE_ASSEMBLY_GAP.search(seq['nt'])
        while m:
            start, end = m.span()

            gap = OrderedDict()
            gap['type'] = bc.FEATURE_GAP
            gap['sequence'] = seq['id']
            gap['start'] = start + 1
            gap['stop'] = end
            gap['strand'] = bc.STRAND_NA
            gap['length'] = end - start

            gaps.append(gap)
            log.info(
                'seq=%s, start=%i, stop=%i, length=%s',
                gap['sequence'], gap['start'], gap['stop'], gap['length']
            )
            m = RE_ASSEMBLY_GAP.search(seq['nt'], end + 1)
    return gaps
